#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.


from copy import deepcopy
import functools
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.state import State
from numbers import Number
from typing import Any, Dict, List, Optional
import pandas as pd
from datetime import date

from mitosheet.sheet_functions.types import (
    BOOLEAN_SERIES,
    DATETIME_SERIES,
    NUMBER_SERIES,
    STRING_SERIES
)
from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_filter_error
)

# SOME CONSTANTS USED IN THE FILTER STEP ITSELF
FC_EMPTY = 'empty'
FC_NOT_EMPTY = 'not_empty'
SHARED_FILTER_CONDITIONS = [
    FC_EMPTY,
    FC_NOT_EMPTY
]

FC_BOOLEAN_IS_TRUE = 'boolean_is_true'
FC_BOOLEAN_IS_FALSE = 'boolean_is_false'
BOOLEAN_FILTER_CONDITIONS = [
    FC_BOOLEAN_IS_TRUE,
    FC_BOOLEAN_IS_FALSE
]

FC_STRING_CONTAINS = 'contains'
FC_STRING_DOES_NOT_CONTAIN = 'string_does_not_contain'
FC_STRING_EXACTLY = 'string_exactly'
FC_STRING_NOT_EXACTLY = 'string_not_exactly'
STRING_FILTER_CONDITIONS = [
    FC_STRING_CONTAINS,
    FC_STRING_DOES_NOT_CONTAIN,
    FC_STRING_EXACTLY,
    FC_STRING_NOT_EXACTLY
]

FC_NUMBER_EXACTLY = 'number_exactly'
FC_NUMBER_NOT_EXACTLY = 'number_not_exactly'
FC_NUMBER_GREATER = 'greater'
FC_NUMBER_GREATER_THAN_OR_EQUAL = 'greater_than_or_equal'
FC_NUMBER_LESS = 'less'
FC_NUMBER_LESS_THAN_OR_EQUAL = 'less_than_or_equal'
NUMBER_FILTER_CONDITIONS = [
    FC_NUMBER_EXACTLY,
    FC_NUMBER_NOT_EXACTLY,
    FC_NUMBER_GREATER,
    FC_NUMBER_GREATER_THAN_OR_EQUAL,
    FC_NUMBER_LESS,
    FC_NUMBER_LESS_THAN_OR_EQUAL
]

FC_DATETIME_EXACTLY = 'datetime_exactly'
FC_DATETIME_GREATER = 'datetime_greater'
FC_DATETIME_GREATER_THAN_OR_EQUAL = 'datetime_greater_than_or_equal'
FC_DATETIME_LESS = 'datetime_less'
FC_DATETIME_LESS_THAN_OR_EQUAL = 'datetime_less_than_or_equal'
DATETIME_FILTER_CONDITIONS = [
    FC_DATETIME_EXACTLY,
    FC_DATETIME_GREATER,
    FC_DATETIME_GREATER_THAN_OR_EQUAL,
    FC_DATETIME_LESS,
    FC_DATETIME_LESS_THAN_OR_EQUAL,
]

class FilterStepPerformer(StepPerformer):
    """
    Allows you to filter a column based on some conditions and some values. 
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'filter_column'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Filtered a Column'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'filter_column_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        """
        Saturates the filter event with a `has_non_empty_filter` - which is useful
        for for logging
        """
        has_non_empty_filter = False
        for filter_or_group in params['filters']:
            if 'filters' in filter_or_group:
                # If it's a group
                if len(filter_or_group['filters']) > 0:
                    has_non_empty_filter = True
            else:
                # If it's a single filter
                has_non_empty_filter = True

        params['has_non_empty_filter'] = has_non_empty_filter
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        column_header: str,
        operator: str,
        filters,
        **params
    ) -> State:
        # if the sheet doesn't exist, throw an error
        if not prev_state.does_sheet_index_exist_within_state(sheet_index):
            raise make_no_sheet_error(sheet_index)

        # We check that the filtered column exists 
        missing_column = set([column_header]).difference(prev_state.column_metatype[sheet_index].keys())
        if len(missing_column) > 0: 
            raise make_no_column_error(missing_column)

        # If no errors we create a new step for this filter
        post_state = deepcopy(prev_state)

        # Execute the filter
        post_state.dfs[sheet_index] = _execute_filter(
            prev_state.dfs[sheet_index], 
            column_header,
            operator,
            filters
        )

        # keep track of which columns are filtered
        post_state.column_filters[sheet_index][column_header]['operator'] = operator
        post_state.column_filters[sheet_index][column_header]['filters'] = filters

        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        column_header: str,
        operator: str,
        filters,
        **params
    ) -> List[str]:
        df_name = post_state.df_names[sheet_index]

        filter_strings = []
        for filter_or_group in filters:
            # If it is a group, we build the code for each filter, and then combine them at the end
            if 'filters' in filter_or_group:
                group_filter_strings = []
                for filter_ in filter_or_group['filters']:
                    group_filter_strings.append(
                        get_filter_string(df_name, column_header, filter_)
                    )

                filter_strings.append(
                    # Note: we add parens around this, so it's grouped properly
                    "(" + combine_filter_strings(filter_or_group['operator'], group_filter_strings) + ")"
                )
            else:
                filter_strings.append(
                    get_filter_string(df_name, column_header, filter_or_group)
                )

        if len(filter_strings) == 0:
            return []
        elif len(filter_strings) == 1:
            return [
                f'{df_name} = {df_name}[{filter_strings[0]}]',
                f'{df_name} = {df_name}.reset_index(drop=True)'
            ]
        else:
            filter_string = combine_filter_strings(operator, filter_strings, split_lines=True)
            return [
                f'{df_name} = {df_name}[{filter_string}]',
                f'{df_name} = {df_name}.reset_index(drop=True)'
            ]

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        column_header: str,
        operator: str,
        filters,
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            df_name = df_names[sheet_index]
            return f'Filtered {column_header} in {df_name}'
        return f'Filtered {column_header}'


def get_applied_filter(df, column_header, filter_):
    """
    Given a filter triple, returns the filter indexes for that
    actual dataframe
    """
    type_ = filter_['type']
    condition = filter_['condition']
    value = filter_['value']

    # First, we case on the shared conditions, to get them out of the way
    if condition in SHARED_FILTER_CONDITIONS:
        if condition == FC_EMPTY:
            return df[column_header].isna()
        elif condition == FC_NOT_EMPTY:
            return df[column_header].notnull()
    
    if type_ == BOOLEAN_SERIES:
        if condition not in BOOLEAN_FILTER_CONDITIONS:
            raise Exception(f'Invalid condition passed to boolean filter {condition}')

        if condition == FC_BOOLEAN_IS_TRUE:
            return df[column_header] == True
        elif condition == FC_BOOLEAN_IS_FALSE:
            return df[column_header] == False
    elif type_ == STRING_SERIES:
        if condition not in STRING_FILTER_CONDITIONS:
            raise Exception(f'Invalid condition passed to string filter {condition}')

        # Check that the value is the valid
        if not isinstance(value, str):
            raise make_invalid_filter_error(value, STRING_SERIES)

        if condition == FC_STRING_CONTAINS:
            return df[column_header].str.contains(value, na=False)
        if condition == FC_STRING_DOES_NOT_CONTAIN:
            return ~df[column_header].str.contains(value, na=False)
        elif condition == FC_STRING_EXACTLY:
            return df[column_header] == value
        elif condition == FC_STRING_NOT_EXACTLY:
            return df[column_header] != value

    elif type_ == NUMBER_SERIES:
        if condition not in NUMBER_FILTER_CONDITIONS:
            raise Exception(f'Invalid condition passed to number filter {condition}')
        
        # Check that the value is the valid
        if not isinstance(value, Number):
            raise make_invalid_filter_error(value, NUMBER_SERIES)

        if condition == FC_NUMBER_EXACTLY:
            return df[column_header] == value
        elif condition == FC_NUMBER_NOT_EXACTLY:
            return df[column_header] != value
        elif condition == FC_NUMBER_GREATER:
            return df[column_header] > value
        elif condition == FC_NUMBER_GREATER_THAN_OR_EQUAL:
            return df[column_header] >= value
        elif condition == FC_NUMBER_LESS:
            return df[column_header] < value
        elif condition == FC_NUMBER_LESS_THAN_OR_EQUAL:
            return df[column_header] <= value

    elif type_ == DATETIME_SERIES:
        if condition not in DATETIME_FILTER_CONDITIONS:
            raise Exception(f'Invalid condition passed to datetime filter {condition}')

        # Check that we were given something that can be understood as a date
        try:
            timestamp = pd.to_datetime(value)
        except:
            # If we hit an error, because we restrict the input datetime, 
            # this is probably occuring because the user has only partially input the date, 
            # and so in this case, we just default it to the minimum possible timestamp for now!
            timestamp = date.min

        if condition == FC_DATETIME_EXACTLY:
            return df[column_header] == timestamp
        elif condition == FC_DATETIME_GREATER:
            return df[column_header] > timestamp
        elif condition == FC_DATETIME_GREATER_THAN_OR_EQUAL:
            return df[column_header] >= timestamp
        elif condition == FC_DATETIME_LESS:
            return df[column_header] < timestamp
        elif condition == FC_DATETIME_LESS_THAN_OR_EQUAL:
            return df[column_header] <= timestamp
    else:
        raise Exception(f'Invalid type passed in filter {type_}')

def combine_filters(operator, filters):

    def filter_reducer(filter_one, filter_two):
        # helper for combining filters based on the operations
        if operator == 'Or':
            return (filter_one) | (filter_two)
        elif operator == 'And':
            return (filter_one) & (filter_two)
        else:
            raise Exception(f'Operator {operator} is unsupported')

    # Combine all the filters into a single filter
    return functools.reduce(filter_reducer, filters)

def _execute_filter(
        df, 
        column_header,
        operator,
        filters
    ):
    """
    Executes a filter on the given column, filtering by removing any rows who
    don't meet the condition.
    """

    applied_filters = []

    for filter_or_group in filters:

        # If it's a group, then we build the filters for the group, combine them
        # and then add that to the applied filters
        if 'filters' in filter_or_group:
            group_filters = []
            for filter_ in filter_or_group['filters']:
                group_filters.append(
                    get_applied_filter(df, column_header, filter_)
                )

            if len(group_filters) > 0:
                applied_filters.append(
                    combine_filters(filter_or_group['operator'], group_filters)
                )    

        # Otherwise, we just get that specific filter, and append it
        else:
            applied_filters.append(get_applied_filter(df, column_header, filter_or_group))    
    
    if len(applied_filters) > 0:
        return df[combine_filters(operator, applied_filters)].reset_index(drop=True)
    else:
        return df


def get_filter_string(df_name, column_header, filter_):
    """
    Transpiles a specific filter to a fitler string, to be used
    in constructing the final transpiled code
    """
    condition = filter_['condition']
    value = filter_['value']

    FILTER_FORMAT_STRING_DICT = {
        # SHARED CONDITIONS
        FC_EMPTY: '{df_name}.{column_header}.isna()',
        FC_NOT_EMPTY: '{df_name}.{column_header}.notnull()',

        # BOOLEAN
        FC_BOOLEAN_IS_TRUE: '{df_name}[\'{column_header}\'] == True',
        FC_BOOLEAN_IS_FALSE: '{df_name}[\'{column_header}\'] == False',

        # NUMBERS
        FC_NUMBER_EXACTLY: '{df_name}[\'{column_header}\'] == {value}',
        FC_NUMBER_NOT_EXACTLY: '{df_name}[\'{column_header}\'] != {value}',
        FC_NUMBER_GREATER: '{df_name}[\'{column_header}\'] > {value}',
        FC_NUMBER_GREATER_THAN_OR_EQUAL: '{df_name}[\'{column_header}\'] >= {value}',
        FC_NUMBER_LESS: '{df_name}[\'{column_header}\'] < {value}',
        FC_NUMBER_LESS_THAN_OR_EQUAL: '{df_name}[\'{column_header}\'] <= {value}',
        
        # STRINGS
        FC_STRING_CONTAINS: '{df_name}[\'{column_header}\'].str.contains(\'{value}\', na=False)',
        FC_STRING_DOES_NOT_CONTAIN: '~{df_name}[\'{column_header}\'].str.contains(\'{value}\', na=False)',
        FC_STRING_EXACTLY: '{df_name}[\'{column_header}\'] == \'{value}\'',
        FC_STRING_NOT_EXACTLY: '{df_name}[\'{column_header}\'] != \'{value}\'',

        # DATES
        FC_DATETIME_EXACTLY: '{df_name}[\'{column_header}\'] == pd.to_datetime(\'{value}\')',
        FC_DATETIME_GREATER: '{df_name}[\'{column_header}\'] > pd.to_datetime(\'{value}\')',
        FC_DATETIME_GREATER_THAN_OR_EQUAL: '{df_name}[\'{column_header}\'] >= pd.to_datetime(\'{value}\')',
        FC_DATETIME_LESS: '{df_name}[\'{column_header}\'] < pd.to_datetime(\'{value}\')',
        FC_DATETIME_LESS_THAN_OR_EQUAL: '{df_name}[\'{column_header}\'] <= pd.to_datetime(\'{value}\')',            
    }

    return FILTER_FORMAT_STRING_DICT[condition].format(
        df_name=df_name,
        column_header=column_header,
        value=value
    )

def combine_filter_strings(operator, filter_strings, split_lines=False):
    """
    Combines the given filter strings with the passed operator, optionally 
    splitting the lines at 120 characters.
    
    NOTE: we choose to keep groups together for readibility, and so do not
    split the lines if we are combing a group.
    """
    if len(filter_strings) == 1:
        return filter_strings[0]
    else:
        # If there are multiple conditions, we combine them together, with the
        # given operator in the middle
        OPERATOR_SIGNS = {
            'Or': '|',
            'And': '&'
        }
        # Put parens around them
        filter_strings = [
            f'({fs})' for fs in filter_strings
        ]

        filter_string = ''
        current_line_length = 0
        for i, fs in enumerate(filter_strings):
            if i != 0:
                filter_string += f' {OPERATOR_SIGNS[operator]} '
            filter_string += fs
            # We keep track of how long the lines are, and if they go over 100 characters,
            # then we split them into a new line (not if this is the last one though)
            current_line_length += len(fs)
            if split_lines and current_line_length > 100 and i != len(filter_strings) - 1:
                filter_string += ' \\\n\t'
                current_line_length = 0

        return filter_string


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Preprocessing occurs to data that is imported into the sheet, and occurs
to _any_ data that is input into the mitosheet - they always occur, and
require no user input. 

For this reason, we take special care to preserve the backwards 
compatability of this step - as all future steps may be reliant on
transformations that have bene made.  

Each pre-processing step has an execution function, and a 
transpilation function, and are run in the same order as
they appear in the list below.
"""

from mitosheet.preprocessing.preprocess_read_file_paths import PREPROCESS_READ_FILE_PATH
from mitosheet.preprocessing.preprocess_rename_headers import PREPROCESS_RENAME_HEADERS
from mitosheet.preprocessing.preprocess_check_args_type import PREPROCESS_CHECK_ARGS_TYPE
from mitosheet.preprocessing.preprocess_copy import PREPROCESS_COPY

# NOTE: these should be in the order you want to apply them to the arguments
PREPROCESS_STEPS = [
   # First, we make sure all the args are the right type
   PREPROCESS_CHECK_ARGS_TYPE,
   # Then, we copy the args to make sure we don't change them accidently
   PREPROCESS_COPY,
   # First, we read in the files
   PREPROCESS_READ_FILE_PATH,
   # Then, we fix up their headers
   PREPROCESS_RENAME_HEADERS
]




from typing import List
import pandas as pd
import plotly.graph_objects as go

from mitosheet.mito_analytics import log
from mitosheet.api.graph.graph_utils import CREATE_FIG_CODE, HISTOGRAM, SHOW_FIG_CODE, X, get_graph_title


def get_histogram(axis, df, column_headers):
    """
    Returns a histogram using the following heuristic:

    If any of the series are not NUMBER_SERIES, return a blank graph.

    Numeric histograms don't require any filtering.  
    """
    fig = go.Figure()

    for column_header in column_headers:
        if axis == X:
            fig.add_trace(go.Histogram(x=df[column_header], name=column_header))
        else:
            fig.add_trace(go.Histogram(y=df[column_header], name=column_header))

    if len(column_headers) == 1:
        if axis == X:
            fig.update_layout(
                xaxis_title=column_headers[0]
            )
        else:
            fig.update_layout(
                yaxis_title=column_headers[0]
            )

    graph_title = get_graph_title(column_headers, [], False, HISTOGRAM, 'frequencies')
    fig.update_layout(
        title=graph_title,
        barmode='group'
    )

    log(f'generate_graph', {
        'params_graph_type': HISTOGRAM,
        'params_axis': axis,
        'params_column_headers': column_headers,
    })

    return fig, False

def get_histogram_code(
        axis: str, 
        df: pd.DataFrame, 
        column_headers: List[str], 
        df_name: str
    ):
    """
    Generates code for a histogram, as above.
    """

    graph_title = get_graph_title(column_headers, [], False, HISTOGRAM, 'frequencies')

    # Note: we conditionally include this code
    update_layout_code = ''
    if len(column_headers) == 1:
        if axis == X:
            update_layout_code = "\n\txaxis_title='{column_header}',".format(column_header=column_headers[0])
        else:
            update_layout_code = "\n\tyaxis_title='{column_header}',".format(column_header=column_headers[0])

    return """{CREATE_FIG_CODE}

# Add the histogram traces to the figure
for column_header in {column_headers}:
    fig.add_trace(go.Histogram({axis}={df_name}[column_header], name=column_header))

# Update the layout
# See Plotly documentation for customizations: https://plotly.com/python/reference/histogram/
fig.update_layout({update_layout_code}
    title='{graph_title}',
    barmode='group'
)
{SHOW_FIG_CODE}""".format(
    CREATE_FIG_CODE=CREATE_FIG_CODE,
    column_headers=column_headers,
    df_name=df_name,
    axis=axis,
    graph_title=graph_title,
    update_layout_code=update_layout_code,
    SHOW_FIG_CODE=SHOW_FIG_CODE
)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.
import os
import pandas as pd
import pytest

from mitosheet.mito_widget import MitoWidget, sheet
from mitosheet.utils import get_invalid_headers
from mitosheet.transpiler.transpile import transpile


def test_example_creation_blank():
    df = pd.DataFrame()
    w = MitoWidget(df, tutorial_mode=False)

VALID_DATAFRAMES = [
    (pd.DataFrame()),
    (pd.DataFrame(data={'A': [1, 2, 3]})),
    (pd.DataFrame(data={'A0123': [1, 2, 3]})),
]
@pytest.mark.parametrize("df", VALID_DATAFRAMES)
def test_sheet_creates_valid_dataframe(df):
    mito = sheet(df)
    assert mito is not None

NON_VALID_HEADER_DATAFRAMES = [
    (pd.DataFrame(data={0: [1, 2, 3]}), ['0_']),
    (pd.DataFrame(data={0.1: [1, 2, 3]}), ['0_1_']),
    (pd.DataFrame(data={'A': [1, 2, 3], 0: [1, 2, 3]}), ['A', '0_']),
    (pd.DataFrame(data={'A': [1, 2, 3], '000': [1, 2, 3]}), ['A', '000_']),
    (pd.DataFrame(data={'A': [1, 2, 3], 'abc-123': [1, 2, 3]}), ['A', 'abc_123']),
    (pd.DataFrame(data={'A': [1, 2, 3], '-123': [1, 2, 3]}), ['A', '_123']),
    (pd.DataFrame(data={'A': [1, 2, 3], '-123_': [1, 2, 3]}), ['A', '_123_']),
    (pd.DataFrame(data={'A': [1, 2, 3], '-abc_': [1, 2, 3]}), ['A', '_abc_']),
    (pd.DataFrame(data={'123': [1, 2, 3], '-abc_': [1, 2, 3]}), ['123_', '_abc_']),
    (pd.DataFrame(data={'ABC!DEF': [1, 2, 3], '123': [1, 2, 3]}), ['ABC_DEF', '123_']),
    (pd.DataFrame(data={'ABC?DEF': [1, 2, 3], '123': [1, 2, 3]}), ['ABC_DEF', '123_']),
    (pd.DataFrame(data={' ': [1, 2, 3], '123': [1, 2, 3]}), ['_', '123_']),
    (pd.DataFrame(data={'-': [1, 2, 3], ' !': [1, 2, 3]}), ['_', '__']),
    (pd.DataFrame(data={'##': [1, 2, 3], ' !': [1, 2, 3]}), ['numnum', '__']),
    (pd.DataFrame(data={'#!?5': [1, 2, 3], '#123': [1, 2, 3]}), ['num__5', 'num123']),
    (pd.DataFrame(data={'()': [1, 2, 3], '.,,': [1, 2, 3]}), ['__', '___']),
    (pd.DataFrame(data={'nate': [1, 2, 3], '.,,': [1, 2, 3]}), ['nate', '___']),
    (pd.DataFrame(data={'//': [1, 2, 3], '-': [1, 2, 3]}), ['__', '_']),
]

@pytest.mark.parametrize("df, new_keys", NON_VALID_HEADER_DATAFRAMES)
def test_sheet_errors_during_non_string_headers(df, new_keys):
    assert len(get_invalid_headers(df)) != 0
    mito = sheet(df)
    assert list(mito.steps_manager.curr_step.dfs[0].keys()) == new_keys

def test_create_with_multiple_dataframes():
    mito = sheet(pd.DataFrame(), pd.DataFrame(), pd.DataFrame(), pd.DataFrame(data={'A': [1, 2, 3]}))
    assert mito is not None

def test_errors_on_non_existing_analysis():
    with pytest.raises(ValueError):
        mito = sheet(saved_analysis_name='bork a dork')


def test_can_call_sheet_with_just_filename():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})
    df.to_csv('../1.csv', index=False)

    mito = sheet('../1.csv')

    assert len(mito.steps_manager.dfs) == 1
    assert mito.steps_manager.dfs[0].equals(df)

    # Remove the test file
    os.remove('../1.csv')


def test_can_call_sheet_with_filename_mulitiple_times():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})
    df.to_csv('../1.csv', index=False)

    mito = sheet('../1.csv', '../1.csv')

    assert len(mito.steps_manager.dfs) == 2
    assert mito.steps_manager.dfs[0].equals(df)
    assert mito.steps_manager.dfs[1].equals(df)
    assert mito.steps_manager.curr_step.df_names[0] != mito.steps_manager.curr_step.df_names[1]

    # Remove the test file
    os.remove('../1.csv')


def test_can_call_sheet_with_df_and_filename():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})
    df.to_csv('../1.csv', index=False)

    mito = sheet(df, '../1.csv')

    assert len(mito.steps_manager.dfs) == 2
    assert mito.steps_manager.dfs[0].equals(mito.steps_manager.dfs[1])

    # Remove the test file
    os.remove('../1.csv')

    code_container = transpile(mito.steps_manager)

    assert code_container['code'] == [
        '# Read in filepaths as dataframes',
        'df_1_csv = pd.read_csv(r\'../1.csv\')',
    ]


def test_call_makes_copies():
    df = pd.DataFrame(data={'A A': [1, 2, 3]})
    sheet(df)

    # Test we don't change the headers!
    assert df.columns.tolist() == ['A A']

def test_can_call_with_indexes():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': ['A', 'B', 'C'], 'D': ['E', 'F', 'G']})

    string_index = df.set_index('B')
    sheet(string_index)

    multi_index = df.set_index(['B', 'D'])
    sheet(multi_index)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Replays an existing analysis onto the sheet
"""

from mitosheet.saved_analyses import read_and_upgrade_analysis


REPLAY_ANALYSIS_UPDATE_EVENT = 'replay_analysis_update'
REPLAY_ANALYSIS_UPDATE_PARAMS = [
    'analysis_name',
    'import_summaries',
    'clear_existing_analysis',
]

def execute_replay_analysis_update(
        steps_manager,
        analysis_name,
        import_summaries,
        clear_existing_analysis
    ):
    """
    This function reapplies all the steps summarized in the passed step summaries, 
    which come from a saved analysis. 

    If any of the step summaries fails, this function tries to roll back to before
    it applied any of the stems

    If clear_existing_analysis is set to true, then this will clear the entire widget
    state container (except the initalize step) before applying the saved analysis.
    """ 
    # We only keep the intialize step only
    if clear_existing_analysis:
        steps_manager.steps = steps_manager.steps[:1]

    # If we're getting an event telling us to update, we read in the steps from the file
    analysis = read_and_upgrade_analysis(analysis_name)

    # When replaying an analysis with import events, you can also send over
    # new params to the import events to replace them. We replace them in the steps here
    if import_summaries is not None:
        for step_idx, file_names in import_summaries.items():
            # NOTE: we have to parse the step index, as it is a string (as it is
            # sent in JSON, which only has strings as keys
            analysis['steps_data'][int(step_idx)]['params']['file_names'] = file_names

    # We stupidly store our saved steps here as a mapping, so we go through and turn it into 
    # a list so that we can pass it into other functions
    steps_manager.execute_steps_data(new_steps_data=analysis['steps_data'])


REPLAY_ANALYSIS_UPDATE = {
    'event_type': REPLAY_ANALYSIS_UPDATE_EVENT,
    'params': REPLAY_ANALYSIS_UPDATE_PARAMS,
    'execute': execute_replay_analysis_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.


from mitosheet.saved_analyses.save_utils import (
    read_and_upgrade_analysis, write_analysis, saved_analysis_names_json,
    delete_saved_analysis, rename_saved_analysis, register_analysis,
    _get_all_analysis_filenames, _delete_analyses,
    SAVED_ANALYSIS_FOLDER, write_analysis, read_analysis, read_saved_analysis_names,
    make_steps_json_obj
)
from mitosheet.saved_analyses.upgrade import is_prev_version
from mitosheet.saved_analyses.upgrade import upgrade_saved_analysis_to_current_version



import io
from typing import List
import pandas as pd

from mitosheet.sheet_functions.types.utils import get_mito_type, NUMBER_SERIES

# We have a variety of heuristics to make sure that we never send too much data
# to the frontend to display to the user. See comments below in the file for 
# which heuristics we use. They use the following constants 

# Graph types should be kept consistent with the GraphType in GraphSidebar.tsx
SCATTER = 'scatter'
BAR = 'bar'
HISTOGRAM = 'histogram'
BOX = 'box'
SUMMARY_STAT = 'summary_stat'

# Max number of unique non-number items to display in a graph
# NOTE: make sure to change both in unison so they make sense
MAX_UNIQUE_NON_NUMBER_VALUES = 10_000
MAX_UNIQUE_NON_NUMBER_VALUES_COMMENT = '(Top 10k)'

SCATTER_PLOT_LESS_THAN_4_SERIES_MAX_NUMBER_OF_ROWS = 10_000
SCATTER_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS = 5_000
BAR_CHART_MAX_NUMBER_OF_ROWS = 5_000
BOX_PLOT_3_SERIES_MAX_NUMBER_OF_ROWS = 500_000
BOX_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS = 250_000

# On the key axis, the difference between the minimum value and the maximum
# value initially displayed on the graph can be no larger than STARTING_RANGE_RANGE
# for qualifying graphs. Only scatter plots and bar charts with a key series that is 
# a number have the default starting range updated. 
STARTING_RANGE_MAX_RANGE = 500

# Labels used to show that the graph is filtered. 
# Used both in the graph_title and the title.
# Note: Because box plot filtering is more complicated, we don't 
# have a graph_filter_label for it
GRAPH_FILTER_LABELS = {
    SCATTER: '(first 10k)',
    BAR: '(first 5k)',
    HISTOGRAM: '(all data)'
}

# Label for each type of graph used in the graph title
GRAPH_TITLE_LABELS = {
    SCATTER: 'scatter plot',
    BAR: 'bar chart',
    BOX: 'box plot',
    HISTOGRAM: 'histogram'
}

# Some, uh, constants that are nice
X = 'x'
Y = 'y'

CREATE_FIG_CODE = """# Import plotly and create a figure
import plotly.graph_objects as go
fig = go.Figure()
"""

# We just use fig.show(renderer="iframe"), which per testing works in both JLab 2
# and JLab 3, and renders in line.
# per testing works in JLab 2 as well. 
SHOW_FIG_CODE = 'fig.show(renderer="iframe")'

def is_all_number_series(df, column_headers):
    """
    Returns True if the mito type of each series with the header column_header
    in column_headers is a NUMBER_SERIES. Returns False otherwise. 
    """
    for column_header in column_headers:
        mito_type = get_mito_type(df[column_header])
        if mito_type != NUMBER_SERIES:
            return False
    return True

def get_graph_title(x_axis_column_headers, y_axis_column_headers, filtered, graph_type, special_title=None):
    """
    Helper function for determing the title of the graph for the scatter plot and bar chart
    """
    # Get the label to let the user know that their graph had a filter applied.

    # Handle the special case for the box plot.
    if graph_type == BOX and filtered:
        graph_filter_label = '(top 500k)' if len(x_axis_column_headers + y_axis_column_headers) == 3 else '(top 250k)'
    # Handle the special case for the scatter plot.
    elif graph_type == SCATTER and filtered:
        graph_filter_label = '(first 10k)' if len(x_axis_column_headers + y_axis_column_headers) < 4 else '(first 5k)'
    else:
        graph_filter_label = GRAPH_FILTER_LABELS[graph_type] if filtered else None

    # Compile all of the column headers into one comma separated string
    all_column_headers = (', ').join(x_axis_column_headers + y_axis_column_headers)
    # Get the title of the graph based on the type of graph
    graph_title_label = GRAPH_TITLE_LABELS[graph_type] if special_title is None else special_title
    # Combine all of the non empty graph title components into one list
    graph_title_components = [all_column_headers, graph_filter_label, graph_title_label] if graph_filter_label is not None else [all_column_headers, graph_title_label]
    
    # Return a string with all of the graph_title_components separated by a space  
    return (' ').join(graph_title_components)


def get_graph_labels(x_axis_column_headers, y_axis_column_headers):
    """
    Helper function for determining the x and y axis titles, 
    for the scatter plot and bar chart. 
    """
    if x_axis_column_headers == [] and y_axis_column_headers == []:
        # If no data is provided, don't label the axises
        x_axis_title = ''
        y_axis_title = ''

    elif x_axis_column_headers == [] and y_axis_column_headers != []:
        # Following from the graph generation, if the user only selects a y axis, 
        # then the y axis is the index column and the columns selected are put on the x axis
        x_axis_title = y_axis_column_headers[0] if len(y_axis_column_headers) == 1 else ''
        y_axis_title = 'index'

    elif x_axis_column_headers != [] and y_axis_column_headers == []:
        # Following from the graph generation, if the user only selects a x axis, 
        # then the y axis is the index column
        x_axis_title = 'index'
        y_axis_title = x_axis_column_headers[0] if len(x_axis_column_headers) == 1 else ''

    else: 
        # Only label the axis if there is one column header on the axis. Otherwise, plotly 
        # legend will label the columns
        x_axis_title = x_axis_column_headers[0] if len(x_axis_column_headers) == 1 else ''
        y_axis_title = y_axis_column_headers[0] if len(y_axis_column_headers) == 1 else ''

    return x_axis_title, y_axis_title

def get_starting_ranges(graph_type, df, x_axis_column_headers, y_axis_column_headers):
    """
    Helper function for determing range that the axis should default to on 
    inital render. It finds the smallest value in the key series and sets the range of it
    [smallest value in series, smallest value in series + STARTING_RANGE_RANGE]

    It returns the object {
        X: [start of range, end of range] or [None, None]
        Y: [start of range, end of range] or [None, None]
    }

    At least one of the axises will be [None, None]
    
    In the following conditions, this function will return a range of [None, None] so 
    as to not overwrite the plotly default:
    1. the graph type is not a scatter plot or a bar chart.
    2. the key series is not a NUMBER_SERIES.
    3. the range of the key series is already less than STARTING_RANGE_RANGE.

    This approach is used to avoid graphs rendering where the data is nearly invisible due to the 
    range of the key axis being so large
    """
    if (graph_type != SCATTER and graph_type != BAR) or (x_axis_column_headers == [] and y_axis_column_headers == []):
        # We only need to overwrite the default range in scatter plots and bar charts. 
        # We also don't care about the case where the graph has no data.
        return {
            X: [None, None],
            Y: [None, None]
        } 

    elif len(x_axis_column_headers) == 0 and len(y_axis_column_headers) > 0:
        key_axis = Y
        # Since the user has no way of zooming in/out the y axis, we regretably 
        # don't start the graph zoomed in because they will be stuck there!
        return {
            X: [None, None],
            Y: [None, None]
        } 

    elif len(x_axis_column_headers) > 0 and len(y_axis_column_headers) == 0:
        # In this case the x axis is the key axis and it is the index series. 
        # Because the index series is always a number, we don't have to do the 
        # mito_type_check and we just make sure that there are more than 
        # DEFAULT_RANGE rows. 
        key_axis = X
        return {
            X: [0, STARTING_RANGE_MAX_RANGE] if len(df) > STARTING_RANGE_MAX_RANGE else [None, None],
            Y: [None, None]
        }

    else: 
        key_axis = X if len(x_axis_column_headers) == 1 else Y
        key_series = df[x_axis_column_headers[0]] if key_axis == X else df[y_axis_column_headers[0]]

        # If the series is not a number series, we give up on creating
        # a heuristic for setting the x axis 
        # TODO: Figure out if we can set the range as a list of the first 500 unique values
        # Think about whether the order of the unique values in the chart matters. 
        if get_mito_type(key_series) != NUMBER_SERIES:
            return {
                X: [None, None],
                Y: [None, None]
            } 

        # Learn about the range of the key series to make decision
        # about how we should set the default range of the graph
        min_value_in_series = key_series.min()
        max_value_in_series = key_series.max()

        if max_value_in_series - min_value_in_series > STARTING_RANGE_MAX_RANGE:
            # If the range of the series is greater than STARTING_RANGE_RANGE, then 
            # set the starting range from the min_value_in_series to min_value_in_series + STARTING_RANGE_RANGE
            return {
                X: [min_value_in_series, min_value_in_series + STARTING_RANGE_MAX_RANGE] if key_axis == X else [None, None],
                Y: [None, None] # the Y axis is never set zoomed in because the user has no way of zooming out
            }
        else:
            # If the range of the series is less than STARTING_RANGE_RANGE, 
            # don't overwrite plotly's default range
            return {
                X: [None, None],
                Y: [None, None]
            } 

def filter_df_to_top_unique_values_in_series(
        df: pd.DataFrame,
        main_series: pd.Series,
        num_unique_values: int,
    ) -> pd.Series: 
    """
    Helper function for filtering the dataframe down to the top most common
    num_unique_values in the main_series. Will not change the series if there are less
    values than that.

    The function filters the entire dataframe to make sure that the columns stay 
    the same length (which is necessary if you want to graph them).

    It returns the filtered dataframe
    """
    if len(main_series) < num_unique_values or main_series.nunique() < num_unique_values:
        return df

    value_counts_series = main_series.value_counts()
    most_frequent_values_list = value_counts_series.head(n=num_unique_values).index.tolist()

    return df[main_series.isin(most_frequent_values_list)]

def filter_df_to_safe_size(
        graph_type, 
        df: pd.DataFrame, 
        column_headers: List[str],
        other_axis_column_headers: List[str]=None
    ):
    """
    A helper function that filters a dataframe down to a safe size
    to display in a graph, depending on the type of graph.

    This is used to stop the graph from crashing the users browser
    when it displays too much data.

    It is also exported from the mitosheet package so that it can
    be used in the code that is exported from graphs.
    """
    original_df_len = len(df)

    if graph_type == BOX:
        if len(column_headers) <= 2:
            return df, False
        elif len(column_headers) == 3:
            return df.head(BOX_PLOT_3_SERIES_MAX_NUMBER_OF_ROWS), original_df_len > BOX_PLOT_3_SERIES_MAX_NUMBER_OF_ROWS
        else:
            return df.head(BOX_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS), original_df_len > BOX_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS
    elif graph_type == SCATTER:
        for column_header in column_headers:
            if get_mito_type(df[column_header]) != NUMBER_SERIES:
                # For each non-number series, filter it to only contain the most common values
                df = filter_df_to_top_unique_values_in_series(
                    df, 
                    df[column_header],
                    MAX_UNIQUE_NON_NUMBER_VALUES,
                )

        # Then, take the first 10k rows
        total_allowed_rows = SCATTER_PLOT_LESS_THAN_4_SERIES_MAX_NUMBER_OF_ROWS if len(column_headers) < 4 else SCATTER_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS
        df = df.head(total_allowed_rows)

        return df, len(df) < original_df_len
    elif graph_type == BAR:
        # If both axises are supplied, then we go on to create the bar chart. 
        # Start by sorting the key column in decreasing frequency so we can filter 
        # dataframe according to the bar chart filtering strategy. 
        if other_axis_column_headers is not None and len(other_axis_column_headers) == 1:
            df['Frequency'] = df.groupby(other_axis_column_headers[0])[other_axis_column_headers[0]].transform('count')
        else:
            df['Frequency'] = df.groupby(column_headers[0])[column_headers[0]].transform('count')

        df = df.sort_values('Frequency', inplace=False, ascending=False)

        # Take the top BAR_CHART_MAX_NUMBER_OF_ROWS rows of the sorted dataframe
        df = df.head(BAR_CHART_MAX_NUMBER_OF_ROWS)
        
        return df, len(df) < original_df_len


def get_html_and_script_from_figure(fig, height, width):
    """
    Given a plotly figure, generates HTML from it, and returns
    a dictonary with the div and script for the frontend.

    The plotly HTML generated by the write_html function call is a div with two children:
    1. a div that contains the id for the graph itself
    2. a script that actually builds the graph
    
    Because we have to dynamically execute the script, we split these into two 
    strings, to make them easier to do what we need on the frontend
    """
    # Send the graph back to the frontend
    buffer = io.StringIO()
    fig.write_html(
        buffer,
        full_html=False,
        include_plotlyjs=False,
        default_height=height,
        default_width=width,
    )
    
    original_html = buffer.getvalue()
    # First, we remove the main div, and the resulting whitespace, to just have the children
    original_html = original_html[5:]
    original_html = original_html[:-6]
    original_html = original_html.strip()

    # Then, we split the children into the div, and the script 
    # making sure to remove the script tag (so we can execute it)
    script_start = '<script type=\"text/javascript\">'
    script_end = '</script>'
    split_html = original_html.split(script_start)
    div = split_html[0]
    script = split_html[1][:-len(script_end)]

    return {
        'html': div,
        'script': script
    }


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains function for running a topological sort on a graph with
the representation {node: set(adj nodes)}. Note that all
nodes must be in the graph, even if they have no adj nodes, and
should just have an empty set in this case.
"""
from copy import deepcopy

from .errors import EditError, make_circular_reference_error


def visit(column_evaluation_graph, node, visited, finished_order, visited_loop):
    """
    Recursive helper function for topological sort. Throws an
    circular_reference_error if there is a loop.
    """
    # Mark the node as visited
    visited[node] = True

    # And record we visited it during this tree of calls to visit
    visited_loop.add(node)

    for adj_node in column_evaluation_graph[node]:
        if not visited[adj_node]:
            visit(column_evaluation_graph, adj_node, visited, finished_order, visited_loop)
        elif adj_node in visited_loop:
            # If we have visited this node in this subtree, there is a loop
            raise make_circular_reference_error()

    # Remove so we can visit again from elsewhere
    visited_loop.remove(node)
    # And mark this node as finished
    finished_order.append(node)


def topological_sort_columns(column_evaluation_graph):
    """
    Topologically sorts by DFSing the graph, recording the finish order, and
    then returning nodes in reversed finish order.
    """
    visited = {node: False for node in column_evaluation_graph}
    finish_order = []
    # Visit each node in the graph
    for node in column_evaluation_graph:
        if not visited[node]:
            # Keep track of the nodes visited during this set
            # of recursive calls, so we can detect cycles
            visited_loop = set()
            visit(
                column_evaluation_graph,
                node,
                visited,
                finish_order,
                visited_loop
            )

    # Reverse finish order for DFS == topological sort
    finish_order.reverse()
    return finish_order

def subgraph_from_starting_column_header(column_evaluation_graph, starting_column_header):
    """
    Filters down the column_evaluation_graph to just the nodes that can be reached
    from the starting_point, including the starting_point itself.

    This results in us transpiling less code, as we only transpile code that changes
    after a given starting column changes.
    """
    column_evaluation_subgraph = dict()

    nodes_in_subgraph = set([starting_column_header])
    while len(nodes_in_subgraph) > 0:
        curr_node = nodes_in_subgraph.pop()
        column_evaluation_subgraph[curr_node] = column_evaluation_graph[curr_node]
        nodes_in_subgraph.update(column_evaluation_subgraph[curr_node])

    return column_evaluation_subgraph


def creates_circularity(
        column_evaluation_graph,
        address,
        old_dependencies,
        new_dependencies
    ):
    """
    Given a column_evaluation_graph, checks if removing the
    old_dependencies and adding the new_dependencies to this address
    will introduce a circular reference; without modifing the
    column_evaluation_graph!

    Returns False if there is not a circular reference, and returns
    True if there is a circular reference.
    """
    # Copy, so we don't modify
    _column_evaluation_graph = deepcopy(column_evaluation_graph)

    for old_dependency in old_dependencies:
        _column_evaluation_graph[old_dependency].remove(address)
    for new_dependency in new_dependencies:
        _column_evaluation_graph[new_dependency].add(address)

    try:
        # Errors if there is no toplogical sort possible
        topological_sort_columns(_column_evaluation_graph)

        return False
    except EditError as e:
        # Reports this error
        return True
    except Exception as e:
        # TODO: handle this case with a general error?
        return True


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import deepcopy
from mitosheet.sheet_functions.types.utils import NUMBER_SERIES
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.state import State
from typing import Any, Dict, List, Optional
from mitosheet.errors import (
    make_column_exists_error,
    make_no_sheet_error
)

class AddColumnStepPerformer(StepPerformer):
    """""
    A add_column step, which allows you to add a column to 
    a dataframe. 
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'add_column'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Added a Column'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'add_column_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        column_header: str,
        column_header_index: int,
        **params
    ) -> State:
            
        # if the sheet doesn't exist, throw an error
        if not prev_state.does_sheet_index_exist_within_state(sheet_index):
            raise make_no_sheet_error(sheet_index)

        if column_header in prev_state.column_metatype[sheet_index]:
            raise make_column_exists_error(column_header)

        # We add a new step with the added column
        post_state = deepcopy(prev_state)

        # If the column_header_index is out of range, then make the new column the last column
        if column_header_index < 0 or len(prev_state.dfs[sheet_index].columns) <= column_header_index:
            column_header_index = len(prev_state.dfs[sheet_index].columns)

        # Update the state variables
        post_state.column_metatype[sheet_index][column_header] = 'formula'
        post_state.column_type[sheet_index][column_header] = NUMBER_SERIES
        post_state.column_spreadsheet_code[sheet_index][column_header] = '=0'
        post_state.column_python_code[sheet_index][column_header] = f'df[\'{column_header}\'] = 0'
        post_state.column_evaluation_graph[sheet_index][column_header] = set()
        post_state.column_filters[sheet_index][column_header] = {'operator': 'And', 'filters': []}
            
        # Update the dataframe
        post_state.dfs[sheet_index].insert(column_header_index, column_header, 0)
        
        return post_state, {
            'column_header_index': column_header_index
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        column_header: str,
        column_header_index: int
    ) -> List[str]:

        return [
            f'{post_state.df_names[sheet_index]}.insert({execution_data["column_header_index"]}, \'{column_header}\', 0)'
        ]

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        column_header: str,
        column_header_index: int,
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            df_name = df_names[sheet_index]
            return f'Added column {column_header} to {df_name}'
        return f'Added column {column_header}' 


import json
import os


def get_path_parts(path):
    """
    For a path, returns a list of the path broken down into
    pieces.

    TODO: test this on Windows
    """
    drive, path_and_file = os.path.splitdrive(path)
    path, file = os.path.split(path_and_file)

    folders = []
    # https://stackoverflow.com/questions/3167154/how-to-split-a-dos-path-into-its-components-in-python
    # We take the code from here, which apparently avoids entering an infinite loop
    # as it breaks when path != ''
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        elif path != "":
            folders.append(path)

            break

    folders.reverse()
    return [drive] + folders + [file]


def get_path_contents(event):
    """
    Takes an event with path parts in a list, turns them into
    an actual path, and then sends an API response with those 
    path parts
    """ 
    path_parts = event['path_parts']

    # Join the path and normalize it (note this should be OS independent)
    path = os.path.join(*path_parts)
    path = os.path.normpath(path)

    # We default the path to "." on the frontend, but we replace
    # this with the current directory full path so we can get all
    # the path parts correctly 
    if path == '.':
        path = os.getcwd()

    try:
        # This loop defines these variables, but does nothing with them
        # so we can then return them (which is why we break immediately)
        for (dirpath, dirnames, filenames) in os.walk(path):
            break
    
        # We sort the files so they are alphabetical (ignoring case)
        dirnames = sorted(dirnames, key=str.lower)
        filenames = sorted(filenames, key=str.lower)
    except:
        # If we cannot read the current path, this is a result of the fact
        # that there are permission errors (or something), in which case we
        # just return an empty result
        dirnames = []
        filenames = []
    
    # We then filter out any hidden folders, which we don't want users to be able
    # to see (as they would otherwise). 
    # Linux, Max == starts with "."
    # Windows == "$"
    dirnames = [d for d in dirnames if (not d.startswith('.') and not d.startswith('$'))]

    return json.dumps({
        'path': path,
        'path_parts': get_path_parts(path),
        'files': filenames,
        'directories': dirnames
    })
    
    
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
This preprocessing step is responsible for renaming invalid
column headers to headers mito can handle.
"""
import json
import pandas as pd
from mitosheet.utils import get_header_renames, is_valid_header, make_valid_header

def execute_preprocess_rename_headers(args):
    
    # Note, by the time args get here, they should be all dataframes
    # as the CSV paths should be read in by now
    for sheet_index, df in enumerate(args):
        renames = dict()
        for column_header in df.keys():
            if not is_valid_header(column_header):
                valid_header = make_valid_header(column_header)
                renames[column_header] = valid_header
        
        df.rename(columns=renames, inplace=True)

    return args

def transpile_preprocess_rename_headers(
        steps_manager
    ):
    """
    Transpiles an initial column header rename to Python code! 
    May be empty if there is nothing to rename.
    """
    code = []
    for sheet_index, arg in enumerate(steps_manager.original_args):
        if isinstance(arg, pd.DataFrame):
            column_headers = arg.keys()
            renames = get_header_renames(column_headers)
            
            if len(renames) == 0:
                continue

            code.append(
                f'{steps_manager.curr_step.df_names[sheet_index]}.rename(columns={json.dumps(renames)}, inplace=True)'
            )

    if len(code) > 0:
        code.insert(0, '# Rename headers to make them work with Mito')

    return code

PREPROCESS_RENAME_HEADERS = {
    'execute': execute_preprocess_rename_headers,
    'transpile': transpile_preprocess_rename_headers
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains helpful utility functions
"""
import json
from mitosheet.sheet_functions.types.utils import get_datetime_columns, get_float_columns, get_timedelta_columns
import pandas as pd
import numpy as np
import numbers
import uuid
import re

def make_valid_header(column_header):
    """
    Takes a header, and performs replaces against common characters
    to make the column_header valid!
    """
    # If it's a tuple, we turn it into a string before continuing
    if isinstance(column_header, tuple):
        column_header = '_'.join([str(c) for c in column_header]).strip()

    # If it's just numbers, turn it into a string (with an underscore)
    if isinstance(column_header, numbers.Number):
        return str(column_header).replace('.', '_') + '_'

    # If it's just numbers in a string, add an underscore
    if column_header.isdigit():
        return column_header + "_"

    # Replace common invalid seperators with valid seperators
    replace_dict = {
        ' ': '_',
        '-': '_',
        '(': '_',
        ')': '_',
        '/': '_',
        '#': 'num',
        ',': '_',
        '.': '_',
        '!': '_',
        '?': '_'
    }
    for find, replace in replace_dict.items():
        column_header = column_header.replace(find, replace)
    
    if not is_valid_header(column_header):
        # Because we detect column headers using a word match, any word character counts
        # in a valid character
        pattern = re.compile("\w")

        new_header = ''.join([
            c for c in column_header if pattern.search(c)
        ])
        if not is_valid_header(new_header):
            # And then append an underscore, for good measure, and this should fix it!
            new_header = new_header + '_'

        return new_header
    return column_header

def is_valid_header(column_header):
    """
    A header is valid if It is a string that is made up of all word characters,
    with at least one non numeric char, and has at least one char.

    Valid examples: A, ABC, 012B, 213_bac, 123_123
    Invalid examples: 123, 123!!!, ABC!, 123-123

    This is a result of not being able to distingush column headers from constants
    otherwise, and would not be necessary if we had a column identifier!
    """
    
    # Note the start and end characters in the regex, to make sure it's a full match
    return isinstance(column_header, str) and \
        len(column_header) > 0 and \
        re.compile("^\w+$").search(column_header) and \
        not column_header.isdigit()


def get_invalid_headers(df: pd.DataFrame):
    """
    Given a dataframe, returns a list of all the invalid headers this list has. 
    """
    return [
        header for header in df.columns.tolist()
        if not is_valid_header(header)
    ]

def get_header_renames(column_headers):
    """
    Given a list of column headers, returns a mapping from old, invalid headers to
    new, valid headers. Empty if no renames are necessary.
    """
    renames = dict()
    for column_header in column_headers:
        if not is_valid_header(column_header):
            valid_header = make_valid_header(column_header)
            renames[column_header] = valid_header

    return renames


def dfs_to_json(dfs) -> str:
    return json.dumps([df_to_json_dumpsable(df) for df in dfs])


def df_to_json_dumpsable(df: pd.DataFrame):
    """
    Returns a dataframe represented in a way that can be turned into a 
    JSON object with json.dumps
    """
    # First, we get the head, and then we make a copy (as we modify the df below)
    # NOTE: we get the head first for efficiency reasons.
    df = df.head(n=2000).copy(deep=True) # we only show the first 2k rows!
    
    # Second, we figure out which of the columns contain dates, and we
    # convert them to string columns (for formatting reasons).
    # NOTE: we don't use date_format='iso' in df.to_json call as it appends seconds to the object, 
    # see here: https://stackoverflow.com/questions/52730953/pandas-to-json-output-date-format-in-specific-form
    date_columns = get_datetime_columns(df)
    for column_header in date_columns:
        df[column_header] = df[column_header].dt.strftime('%Y-%m-%d %X')

    # Third, we figure out which of the columns contain timedeltas, and 
    # we format the timedeltas as strings to make them readable
    timedelta_columns = get_timedelta_columns(df)
    for column_header in timedelta_columns:
        df[column_header] = df[column_header].apply(lambda x: str(x))

    # Then, we get all the float columns and actually make them 
    # look like floating point values, by converting them to strings
    float_columns = get_float_columns(df)
    for column_header in float_columns:
        # Convert the value to a string if it is a number, but leave it alone if its a NaN 
        # as to preserve the formatting of NaN values. 
        df[column_header] = df[column_header].apply(lambda x: x if np.isnan(x) else str(x))

    json_obj = json.loads(df.to_json(orient="split"))
    # Then, we go through and find all the null values (which are infinities),
    # and set them to 'NaN' for display in the frontend.
    for d in json_obj['data']:
        for idx, e in enumerate(d):
            if e is None:
                d[idx] = 'NaN'

    return json_obj


def get_random_id():
    """
    Creates a new random ID for the user, which for any given user,
    should only happen once.
    """
    return str(uuid.uuid1())

def get_new_id():
    return str(uuid.uuid4())





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.

import json
import uuid 
from copy import copy, deepcopy
import pandas as pd
from typing import List, Tuple

from mitosheet.step_performers.import_steps.simple_import import SimpleImportStepPerformer
from mitosheet.state import State
from mitosheet.step import Step
from mitosheet.step_performers import EVENT_TYPE_TO_STEP_PERFORMER
from mitosheet.updates import UPDATES
from mitosheet.preprocessing import PREPROCESS_STEPS
from mitosheet.utils import dfs_to_json, get_new_id
from mitosheet.transpiler.transpile import transpile
from mitosheet.user import get_user_field
from mitosheet.data_in_mito import DataTypeInMito, get_data_type_in_mito
from mitosheet.user import (
    UJ_USER_EMAIL, UJ_RECEIVED_TOURS
)


def get_step_indexes_to_skip(step_list: List[Step]):
    """
    Given a list of steps, will collect all of the steps
    from this list that should be skipped.
    """
    step_indexes_to_skip = set()

    for step_index, step in enumerate(step_list):
        step_indexes_to_skip = step_indexes_to_skip.union(
            step.step_indexes_to_skip(step_list[:step_index])
        )
    
    return step_indexes_to_skip

def execute_step_list_from_index(step_list: List[Step], start_index=None):
    """
    Given a list of steps, and a specific index to start from, will assume that 
    the step_list[start_index] is valid, and execute this list of steps from 
    there. See find_last_valid_index to see how the start_index is determined.

    Notably, will _not_ execute the steps that are skipped in this list, which
    means that the returned step list will only have valid prev_state/post_states 
    for the steps that are not skipped.

    If start_index is not given, will start from the initialize step.
    """
    if start_index is None or start_index < 0:
        start_index = 0
    
    # Get the steps to skip, so that we can skip them
    step_indexes_to_skip = get_step_indexes_to_skip(step_list)
    
    # Get the steps that are valid, and the last valid step, so we can execute from there
    new_step_list = step_list[:start_index + 1] 
    last_valid_step = step_list[start_index]


    for partial_index, step in enumerate(step_list[start_index + 1:]):
        step_index = partial_index + start_index + 1
        # If we're skipping a step, add it to the new step list (since we don't
        # want to lose it), but don't reexecute it 
        if step_index in step_indexes_to_skip:
            new_step_list.append(step)
            continue
        
        # Create a new step with the same params
        new_step = Step(
            step.step_type,
            step.step_id,
            step.params
        )

        # Set the previous state of the new step, and then update 
        # what the last valid step is
        new_step.set_prev_state_and_execute(last_valid_step.post_state)
        last_valid_step = new_step

        new_step_list.append(new_step)
    
    return new_step_list

class StepsManager():
    """
    The StepsManager holds the list of the steps, and makes sure 
    they are updated properly when processing new edit and update 
    events.

    To see how this happens, consider the full loop of what happens
    when a new edit event is received.

    1.  The StepsManager receives this new event. It creates a new
        step of the right type, with the params from the new event. 
        This is done in the handle_edit_event function.
    2.  Then, the StepsManager tries to run all the steps that exist
        including this new step _that are not currently skipped_. 
        It does so with the execute_and_update_steps function, which heavily 
        relies on the execute_step_list_from_index helper function above.
    3.  If the new step that is created is a step that does overwrite an
        existing step, then the execute_and_update_steps will detect 
        this, and when running all the steps, will skip running the skipped step. 

    In this way, the StepsManager keeps a list of all the steps that were ever
    valid. Note that some part of these steps is "append-only" in spirit, in 
    that they keep the same step type and params. 

    However, as some steps end up getting skipped, it may contain steps that 
    are now invalid or out of date (e.g. an old filter step that got skipped). 

    Furthermore, the steps objects themselves change, just the step type
    and parameters stay the same and are append-only.
    """

    def __init__(self, args):
        """
        When initalizing the StepsManager, we also do preprocessing
        of the arguments that were passed to the mitosheet. 

        All preprocessing can be found in mitosheet/preprocessing, and each of 
        the transformations are applied before the data is considered imported.
        """
        # We just randomly generate analysis names. 
        # We append a UUID to note that this is not an analysis the user has saved.
        self.analysis_name = 'UUID-' + str(uuid.uuid4())

        # The args are a tuple of dataframes or strings, and we start by making them
        # into a list, and making copies of them for safe keeping
        self.original_args = [
            arg.copy(deep=True) if isinstance(arg, pd.DataFrame) else deepcopy(arg) for arg in args
        ]

        # Then, we go through the process of actually preprocessing the args
        for PREPROCESS_STEP in PREPROCESS_STEPS:
            args = PREPROCESS_STEP['execute'](args)

        # Then we initialize the analysis with just a simple initialize step
        self.steps: List[Step] = [
            Step(
                'initialize',
                'initialize',
                {},
                None,
                State(args),
                {}
            )
        ]

        """
        To help with redo, we store a list of a list of the steps that 
        existed in the step manager before the user clicked undo or reset,
        along with the type of operation
        
        An example of what this object contains:
        1.  The user passes a dataframe df1, adds column A to df1, imports df2, and then
            adds B and C to df2. steps=[add A to df1, import df2, add B to df2, add C to df2]
        2.  The user presses undo. Thus, steps=[add A to df1, import df2, add B to df2],
            and undone_step_list_store = [('undo', [add C to df2])].
        3.  The user then presses clear. Thus, steps = [import df2], and 
            undone_step_list_store = [('undo', [add C to df2]), ('clear', [add A to df1, import df2, add B to df2])]
        
        Note that for a undo, we only store the last step that has been undone, but for a clear, 
        we store the entire step list that we are replacing. This makes it possible
        to easily undo the clear after it's done, as we have the proper step list
        around.
        """
        self.undone_step_list_store: List[Tuple[str, List[Step]]] = []

        # We display the state that exists after the curr_step_idx is applied,
        # which means you can never see before the initalize step
        self.curr_step_idx = 0


    @property
    def curr_step(self) -> Step:
        """
        Returns the current step object as a property of the object,
        so reference it with self.curr_step
        """
        return self.steps[self.curr_step_idx]

    @property
    def num_sheets(self) -> int:
        """
        Duh. :)
        """
        return len(self.steps[self.curr_step_idx].dfs)

    @property
    def dfs(self) -> List[pd.DataFrame]:
        return self.steps[self.curr_step_idx].dfs

    @property
    def df_names_json(self) -> str:
        """
        A JSON object containing the names of the dataframes
        """
        return json.dumps({'df_names': self.curr_step.df_names})
    
    @property
    def df_sources_json(self) -> str:
        """
        A JSON object containing the names of the dataframes
        """
        return json.dumps(self.curr_step.df_sources)

    @property
    def sheet_json(self) -> str:
        """
        sheet_json contains a serialized representation of the data
        frames that is then fed into the ag-grid in the front-end. 

        NOTE: we only display the _first_ 2,000 rows of the dataframe
        for speed reasons. This results in way less data getting 
        passed around
        """
        return dfs_to_json(self.curr_step.dfs)
    
    @property
    def df_shape_json(self) -> str:
        """
        Returns the df shape (rows, columns) of each dataframe in the 
        current step!
        """
        return json.dumps([
            {'rows': df.shape[0], 'cols': df.shape[1]}
            for df in self.curr_step.dfs
        ])

    @property
    def column_spreadsheet_code_json(self) -> str:
        """
        column_spreadsheet_code_json is a list of all the spreadsheet
        formulas that users have used, for each sheet they have. 
        """
        return json.dumps(self.curr_step.column_spreadsheet_code)

    @property
    def code_json(self) -> str:
        """
        This code json string is sent to the front-end and is what
        ends up getting displayed in the codeblock. 
        """
        return json.dumps(transpile(self))

    @property
    def column_filters_json(self) -> str:
        """
        This column_filters list is used by the front end to display
        the filtered icons in the UI
        """
        return json.dumps(self.curr_step.column_filters)
    
    @property
    def column_type_json(self) -> str:
        """
        Returns a list of JSON objects that hold the type of each
        data in each column.
        """
        return json.dumps(self.curr_step.column_type)

    @property
    def data_type_in_mito(self) -> DataTypeInMito:
        return get_data_type_in_mito(self.dfs)

    @property
    def user_email(self) -> str:
        """
        Returns the user_email from user.json
        """
        return get_user_field(UJ_USER_EMAIL)

    @property
    def received_tours(self) -> str:
        """
        Returns the tours that the user has received
        """
        return json.dumps(get_user_field(UJ_RECEIVED_TOURS))

    @property
    def step_summaries_list_json(self) -> str:
        """
        Returns a json list of step summaries, not including
        the skipped steps
        """
        step_summaries_list = []
        step_indexes_to_skip = get_step_indexes_to_skip(self.steps)
        for index, step in enumerate(self.steps):
            if step.step_type == 'initialize':
                step_summaries_list.append({
                    'step_id': step.step_id,
                    'step_idx': 0,
                    'step_type': step.step_type,
                    'step_display_name': "Created a mitosheet",
                    'step_description': "Created a new mitosheet"
                })
                continue
                
            if index in step_indexes_to_skip:
                continue

            step_summaries_list.append({
                'step_id': step.step_id,
                'step_idx': index,
                'step_type': step.step_type,
                'step_display_name': step.step_performer.step_display_name(),
                'step_description': step.step_performer.describe(
                    df_names=step.df_names,
                    **step.params,
                )
            })

        return json.dumps(step_summaries_list)

    def handle_edit_event(self, edit_event):
        """
        Updates the widget state with a new step that was created
        by the edit_event. Each edit event creates one new step.

        If there is an error in the creation of the new step, this
        function will not create the new invalid step.
        """
        # NOTE: We ignore any edit if we are in a historical state, for now. This is a result
        # of the fact that we don't allow previous editing currently
        if self.curr_step_idx != len(self.steps) - 1:
            return

        step_performer = EVENT_TYPE_TO_STEP_PERFORMER[edit_event['type']]

        # First, we make a new step
        new_step = Step(
            step_performer.step_type(),
            edit_event['step_id'],
            edit_event['params']
        )

        new_steps = self.steps + [new_step]

        self.execute_and_update_steps(new_steps)

        # If we add a new step, then we clear the last_undone_list_store, as
        # you cannot redo something after you make a new edit
        self.undone_step_list_store = []

    def handle_update_event(self, update_event):
        """
        Handles any event that isn't caused by an edit, but instead
        other types of new data coming from the frontend (e.g. the df names 
        or some existing steps).
        """
        for update in UPDATES:
            if update_event['type'] == update['event_type']:
                # Get the params for this event
                params = {key: value for key, value in update_event.items() if key in update['params']}
                # Actually execute this event
                update['execute'](self, **params)
                # And then return
                return

        raise Exception(f'{update_event} is not an update event!')

    def find_last_valid_index(self, new_steps: List[Step]):
        """
        Given the new_steps, this function performs some logic to figure
        out what the last valid index in the steps is (that execution can
        then start from).
        """

        # Currently, we only remove steps in an undo
        if len(new_steps) < len(self.steps):
            # If we are removing steps, then we figure out what skipped steps 
            # we are losing, and run from right before where we are no longer
            # skipped steps
            no_longer_skipped_indexes = set()
            for step_index, removed_step in enumerate(self.steps[len(new_steps):]):
                previous_steps = self.steps[:len(new_steps) + step_index]
                no_longer_skipped_indexes = no_longer_skipped_indexes.union(
                    removed_step.step_indexes_to_skip(previous_steps)
                )

            last_valid_index = min(no_longer_skipped_indexes.union({len(new_steps)})) - 1
        else:
            # Otherwise, if we're adding steps, we figure out which skipped steps
            # we're adding, and run from right before the oldest new skipped step

            # Collect anything that is newly skipped
            newly_skipped_indexes = set()
            for step_index, new_step in enumerate(new_steps[len(self.steps):]):
                previous_steps = new_steps[:len(self.steps) + step_index]
                newly_skipped_indexes = newly_skipped_indexes.union(
                    new_step.step_indexes_to_skip(previous_steps)
                )
            
            # The last valid index is the minimum of the newly skipped things - 1
            # or the last valid step (if nothing is skipped)
            # TODO: we can improve this in the future to remember what it executed last time
            # and so do less work!
            last_valid_index = min(newly_skipped_indexes.union({len(self.steps)})) - 1
        return last_valid_index

    def execute_undo(self):
        """
        This function attempts to undo the most recent step, and if there
        is no most recent step, does nothing.

        It also handles the special case where the last action was clearing
        the analysis, in which case we have the undo actually reset the 
        entire analysis.
        """
        
        # When a user's most recent action is a clear analysis, then the undone_step_list_store
        # will end in an item that says ('clear', [...]).
        # In this case, if they press undo right after clearing, then we assume they probably 
        # want to undo the clear, aka to redo all those steps
        if len(self.undone_step_list_store) > 0:
            if self.undone_step_list_store[-1][0] == 'clear':
                return self.execute_redo()

        # Otherwise, we just undo the most recent step that the user has created
        # if they have created any steps
        if len(self.steps) == 1:
            return 

        new_steps = copy(self.steps)
        undone_step = new_steps.pop()

        self.execute_and_update_steps(new_steps)

        # If this works, then let's add this step to the undo list!
        self.undone_step_list_store.append(
            ('undo', [undone_step])
        )

    def execute_redo(self):
        """
        Executes a redo, which reapplies the most recently undone
        steps if they exist.

        This will not error if there is nothing to redo, it will
        just return.
        """
        if len(self.undone_step_list_store) == 0:
            return 
        
        (undo_or_clear, step_list) = self.undone_step_list_store[-1]
        if undo_or_clear == 'undo':
            # If it's an undo, just apply onto the end
            new_steps = copy(self.steps)
            new_steps.extend(step_list)
            self.execute_and_update_steps(new_steps)

        elif undo_or_clear == 'clear':
            new_steps = step_list
            # Note: since we're breaking the invariant that the steps don't
            # move order, we have to execute from the very start
            self.execute_and_update_steps(new_steps, last_valid_index=0)
    
        # Remove the item we just redid from the undone_step_list_store, so 
        # that we don't redo it again
        self.undone_step_list_store.pop()
    

    def execute_clear(self):
        """
        A clear update, which removes all steps in the analysis
        that are not imports.    
        """
        if len(self.steps) == 1:
            return 

        # Keep only the initalize step and any import steps
        new_steps = [
            step for step in self.steps
            if (
                step.step_type == 'initialize' or \
                step.step_type == SimpleImportStepPerformer.step_type()
            )
        ]

        old_steps = copy(self.steps)

        # We need to set the last_valid_index to 0 as we are changing
        # the step order
        self.execute_and_update_steps(new_steps, last_valid_index=0)

        self.undone_step_list_store.append(
            ('clear', old_steps)
        )


    def execute_and_update_steps(
            self, 
            new_steps: List[Step], 
            last_valid_index=None
        ):
        """
        Given a list of new_steps, runs them from the last valid index,
        based on what is skipped in these new_steps.

        If these new steps result in valid execution, then we set them
        to the step that the step manager stores. 

        If last_valid_index is passed, then this will be used instead of 
        the last_valid_index that could be calculated by the step list.

        So, pass a last_valid_index if you're changing the order of the steps
        in the new_steps array. Otherwise, the step manager can calculate
        the last valid index without help.
        """
        if last_valid_index is None:
            last_valid_index = self.find_last_valid_index(new_steps)
        
        final_steps = execute_step_list_from_index(new_steps, start_index=last_valid_index)
        self.steps = final_steps
        self.curr_step_idx = len(self.steps) - 1

    def execute_steps_data(self, new_steps_data=None):
        """
        Given steps data (e.g. from a saved analysis), will turn
        this data  into steps and try to run them. If any of them
        fail, will take none of the new steps
        """
        new_steps = copy(self.steps)
        for step_data in new_steps_data:
            new_step = Step(
                step_data['step_type'],
                get_new_id(),
                step_data['params']
            )

            new_steps.append(new_step)

        self.execute_and_update_steps(new_steps)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
This file contains helpful functions and classes for testing operations.
"""

from typing import List
from mitosheet.utils import get_new_id
import pandas as pd
from mitosheet.transpiler.transpile import transpile

from mitosheet.mito_widget import sheet, MitoWidget
import json

class MitoWidgetTestWrapper:
    """
    This class adds some simple wrapper functions onto the MitoWidget 
    to make interacting with it easier for testing purposes.

    It allows you to create just the backend piece of Mito, create columns,
    set formulas, and get values to check the result.
    """

    def __init__(self, mito_widget: MitoWidget):
        self.mito_widget = mito_widget

    @property
    def transpiled_code(self):
        # NOTE: we don't add comments to this testing functionality, so that 
        # we don't have to change tests if we update comments
        return transpile(self.mito_widget.steps_manager, add_comments=False)['code']

    def add_column(self, sheet_index: int, column_header: str, column_header_index=-1):
        """
        Adds a column.
        """

        self.mito_widget.receive_message(self.mito_widget, {
            'event': 'edit_event',
            'id': get_new_id(),
            'type': 'add_column_edit',
            'step_id': get_new_id(),
            'params': {
                'sheet_index': sheet_index,
                'column_header': column_header,
                'column_header_index': column_header_index
            }
        })
    
    def set_formula(
            self, 
            formula: str, 
            sheet_index: int,
            column_header: str, 
            add_column=False,
        ):
        """
        Sets the given column to have formula, and optionally
        adds the column if it does not already exist.
        """
        if add_column:
            self.add_column(sheet_index, column_header)

        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'set_column_formula_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_header': column_header,
                    'new_formula': formula,
                }
            }
        )

    def merge_sheets(
            self, 
            how,
            sheet_index_one, 
            merge_key_one, 
            selected_columns_one,
            sheet_index_two, 
            merge_key_two,
            selected_columns_two
        ):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'merge_edit',
                'step_id': get_new_id(),
                'params': {
                    'how': how,
                    'sheet_index_one': sheet_index_one,
                    'merge_key_one': merge_key_one,
                    'selected_columns_one': selected_columns_one,
                    'sheet_index_two': sheet_index_two,
                    'merge_key_two': merge_key_two,
                    'selected_columns_two': selected_columns_two
                }
            }
        )

    def pivot_sheet(
            self, 
            sheet_index, 
            pivot_rows,
            pivot_columns,
            values,
            destination_sheet_index=None,
            step_id=None
        ):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'pivot_edit',
                'step_id': get_new_id() if step_id is None else step_id,
                'params': {
                    'sheet_index': sheet_index,
                    'pivot_rows': pivot_rows,
                    'pivot_columns': pivot_columns,
                    'values': values,
                    'destination_sheet_index': destination_sheet_index
                }
            }
        )

    def filter(
            self, 
            sheet_index, 
            column,
            operator,
            type_,
            condition, 
            value
        ):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'filter_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_header': column,
                    'operator': operator,
                    'filters': [{
                        'type': type_,
                        'condition': condition,
                        'value': value
                    }]
                }
            }
        )
    
    def filters(
            self, 
            sheet_index, 
            column,
            operator,
            filters
        ):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'filter_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_header': column,
                    'operator': operator,
                    'filters': filters
                }
            }
        )
    
    def sort(
            self, 
            sheet_index, 
            column,
            sort_direction
        ):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'sort_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_header': column,
                    'sort_direction': sort_direction
                }
            }
        )

    def reorder_column(
            self, 
            sheet_index, 
            column_header, 
            new_column_index
        ):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'reorder_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_header': column_header,
                    'new_column_index': new_column_index
                }
            }
        )

    def rename_column(self, sheet_index: int, old_column_header: str, new_column_header):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'rename_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'old_column_header': old_column_header,
                    'new_column_header': new_column_header
                }
            }
        )

    def delete_column(self, sheet_index: int, column_header: str):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'delete_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_header': column_header
                }
            }
        )

    def change_column_dtype(self, sheet_index: int, column_header: str, new_dtype: str):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'change_column_dtype_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_header': column_header,
                    'new_dtype': new_dtype
                }
            }
        )

    def simple_import(self, file_names: List[str]):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'simple_import_edit',
                'step_id': get_new_id(),
                'params': {
                    'file_names': file_names
                }
            }
        )

    
    def undo(self):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'undo'
            }
        )
    
    def redo(self):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'redo'
            }
        )

    def clear(self):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'clear'
            }
        )
    
    def save_analysis(self, analysis_name):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'save_analysis_update',
                'analysis_name': analysis_name
            }
        )


    def rename_analysis(self, old_analysis_name, new_analysis_name):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'rename_analysis_update',
                'old_analysis_name': old_analysis_name,
                'new_analysis_name': new_analysis_name
            }
        )

    
    def delete_analysis(self, analysis_name):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'delete_analysis_update',
                'analysis_name': analysis_name
            }
        )

    def delete_dataframe(self, sheet_index):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'dataframe_delete_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index
                }
            }
        )

    def duplicate_dataframe(self, sheet_index):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'dataframe_duplicate_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                }
            }
        )

    def rename_dataframe(self, sheet_index, new_dataframe_name):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'dataframe_rename_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'new_dataframe_name': new_dataframe_name
                }
            }
        )

    def replay_analysis(self, analysis_name, import_summaries=None, clear_existing_analysis=False):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'replay_analysis_update',
                'analysis_name': analysis_name,
                'import_summaries': import_summaries,
                'clear_existing_analysis': clear_existing_analysis
            }
        )

    def checkout_step_by_idx(self, index):
        self.mito_widget.receive_message(
            self.mito_widget,
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'checkout_step_by_idx_update',
                'step_idx': index,
            }
        )

    def get_formula(self, sheet_index: int, column_header: str):
        """
        Gets the formula for a given column. Returns an empty
        string if nothing exists.
        """
        if column_header not in self.mito_widget.column_spreadsheet_code[sheet_index]:
            return ''
        return self.mito_widget.column_spreadsheet_code[sheet_index][column_header]

    def get_value(self, sheet_index: int, column_header: str, row: int):
        """
        Returns a value in a given dataframe at the a given
        index in a column. NOTE: the row is 1 indexed!

        Errors if the value does not exist
        """
        return self.mito_widget.steps_manager.curr_step.dfs[sheet_index].at[row - 1, column_header]

    def get_column(self, sheet_index: int, column_header: str, as_list: bool):
        """
        Returns a series object of the given column, or a list if
        as_list is True. 

        Errors if the column does not exist. 
        """
        if as_list:
            return self.mito_widget.steps_manager.dfs[sheet_index][column_header].tolist()
        return self.mito_widget.steps_manager.dfs[sheet_index][column_header]

def create_mito_wrapper(sheet_one_A_data, sheet_two_A_data=None) -> MitoWidgetTestWrapper:
    """
    Returns a MitoWidgetTestWrapper instance wrapped around a MitoWidget
    that contains just a column A, containing sheet_one_A_data.
    
    If sheet_two_A_data is defined, then also creates a second dataframe
    with column A defined as this as well.
    """
    dfs = [pd.DataFrame(data={'A': sheet_one_A_data})]

    if sheet_two_A_data is not None:
        dfs.append(pd.DataFrame(data={'A': sheet_two_A_data}))

    mito_widget = sheet(*dfs)
    return MitoWidgetTestWrapper(mito_widget)

def create_mito_wrapper_dfs(*args):
    """
    Creates a MitoWidgetTestWrapper with a mito instance with the given
    data frames.
    """
    mito_widget = sheet(*args)
    return MitoWidgetTestWrapper(mito_widget)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Makes sure we don't accidently deploy single threaded code
"""
from mitosheet.api.api import THREADED

def test_multi_threaded():
    assert THREADED



from typing import List
import pandas as pd
import plotly.graph_objects as go

from mitosheet.mito_analytics import log
from mitosheet.sheet_functions.types.utils import NUMBER_SERIES, get_mito_type
from mitosheet.api.graph.graph_utils import BAR, CREATE_FIG_CODE, MAX_UNIQUE_NON_NUMBER_VALUES_COMMENT, SCATTER, SCATTER_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS, SCATTER_PLOT_LESS_THAN_4_SERIES_MAX_NUMBER_OF_ROWS, SHOW_FIG_CODE, X, Y, filter_df_to_safe_size, filter_df_to_top_unique_values_in_series, get_graph_labels, get_graph_title, get_starting_ranges


def get_scatter_plot(df, x_axis_column_headers, y_axis_column_headers):
    """
    Returns a scatter plot using the following heuristic:

    - If a non-number series exists, filter the df to contain the most common MAX_UNIQUE_NON_NUMBER_VALUES
    - Graph the first SCATTER_PLOT_MAX_NUMBER_OF_ROWS rows regardless of the type of the series. 
    """
    fig = go.Figure()
    original_length_of_df = len(df)

    x_axis = len(x_axis_column_headers) != 0
    y_axis = len(y_axis_column_headers) != 0

    # If no data is passed, return a blank graph
    if not x_axis and not y_axis:
        # This should never happen because this check already occurs in the get_graph
        # function. But we leave it here for robustness of the function
        return go.Figure(), False

    # Check the type of each series to appropriately filter 
    all_column_headers = x_axis_column_headers + y_axis_column_headers
    df, filtered = filter_df_to_safe_size(SCATTER, df, all_column_headers)

    # Determine if the dataframe was sorted
    filter_label = ''
    if len(df) < original_length_of_df: 
        filter_label = '(first 10k)' if len(all_column_headers) < 4 else '(first 5k)'
    
    # If only one axis is defined, plotly defaults to using
    # the index column as the other axis. This is also what Google Sheets
    # does, so we do it too.    
    if x_axis and not y_axis:
        # Only an x_axis was provided
        for column_header in x_axis_column_headers:
            fig.add_trace(go.Scatter(
                y=df[column_header], 
                mode='markers',
                name=(' ').join([column_header, filter_label])
            ))
    elif not x_axis and y_axis:
        # Only a y axis was provided
        for column_header in y_axis_column_headers:
            fig.add_trace(go.Scatter(
                x=df[column_header],
                mode='markers',
                name=(' ').join([column_header, filter_label])
            ))
    elif len(y_axis_column_headers) == 1:
        # If the y axis only has one column header, then we use that as the key
        for column_header in x_axis_column_headers:
            fig.add_trace(go.Scatter(
                x=df[column_header], 
                y=df[y_axis_column_headers[0]],
                mode='markers',
                name=(' ').join([column_header, filter_label])
            ))
    else:
        # It should not be possible for both the x axis and y axis 
        # to have more than 1 column. But if it does happen, we default 
        # to using the first column of the x axis as the key
        for column_header in y_axis_column_headers:
            fig.add_trace(go.Scatter(
                x=df[x_axis_column_headers[0]], 
                y=df[column_header],
                mode='markers',
                name=(' ').join([column_header, filter_label])
            ))

    # Update the layout of the graph
    x_axis_title, y_axis_title = get_graph_labels(x_axis_column_headers, y_axis_column_headers)
    graph_title = get_graph_title(x_axis_column_headers, y_axis_column_headers, filtered, SCATTER)
    fig.update_layout(
        xaxis_title=x_axis_title,
        yaxis_title=y_axis_title,
        title=graph_title,
    )

    # Update the starting range of the graph
    ranges = get_starting_ranges(SCATTER, df, x_axis_column_headers, y_axis_column_headers)
    fig.update_layout(
        xaxis=dict(
            range=ranges[X]
        ), 
        yaxis=dict(
            range=ranges[Y]
        )
    )
    # Determine if the graph was zoomed in
    is_zoomed = ranges[X][0] is not None or ranges[X][1] or ranges[Y][0] is not None or ranges[Y][1] is not None

    log(f'generate_graph', {
        'params_graph_type': SCATTER,
        'params_x_axis_column_headers': x_axis_column_headers,
        'params_x_axis_column_types': [get_mito_type(df[column_header]) for column_header in x_axis_column_headers] if x_axis_column_headers is not None else [],
        'params_y_axis_column_headers': y_axis_column_headers,
        'params_y_axis_column_types': [get_mito_type(df[column_header]) for column_header in y_axis_column_headers] if y_axis_column_headers is not None else [],
        'params_filtered': filtered,
        'params_is_zoomed': is_zoomed
    })

    return fig, is_zoomed

def get_scatter_code(
        df: pd.DataFrame, 
        x_axis_column_headers: List[str], 
        y_axis_column_headers: List[str],
        df_name: str
    ):
    """
    Generates code for a scatter plot. Only difference from the graph being generated
    is that it does not zoom by default.
    """

    _, filtered = filter_df_to_safe_size(SCATTER, df, x_axis_column_headers + y_axis_column_headers)

    filtered_code = ''
    if filtered:
        filtered_code = """
# Filter the dataframe so that it doesn't crash the browser with too much data
from mitosheet import filter_df_to_safe_size
{df_name}_filtered, _ = filter_df_to_safe_size('scatter', {df_name}, {column_headers})
""".format(
    df_name=df_name,
    column_headers=x_axis_column_headers + y_axis_column_headers
)
        df_name = f'{df_name}_filtered'



    x_axis = len(x_axis_column_headers) != 0
    y_axis = len(y_axis_column_headers) != 0      

    if x_axis and not y_axis:
        # Only an x_axis was provided
        trace_code = """
# Add the scatter traces to the figure
for column_header in {x_axis_column_headers}:
    fig.add_trace(go.Scatter(
        y={df_name}[column_header], 
        mode='markers',
        name=column_header
    ))
""".format(
    df_name=df_name, 
    x_axis_column_headers=x_axis_column_headers
)
        
    elif not x_axis and y_axis:
                trace_code = """
# Add the scatter traces to the figure
for column_header in {y_axis_column_headers}:
    fig.add_trace(go.Scatter(
        x={df_name}[column_header], 
        mode='markers',
        name=column_header
    ))
""".format(
    df_name=df_name, 
    y_axis_column_headers=y_axis_column_headers
)
    elif len(y_axis_column_headers) == 1:
        trace_code = """
# Add the scatter traces to the figure
for column_header in {x_axis_column_headers}:
    fig.add_trace(go.Scatter(
        x={df_name}[column_header],
        y={df_name}['{y_axis_column_header}'],
        mode='markers',
        name=column_header
    ))
""".format(
    df_name=df_name, 
    x_axis_column_headers=x_axis_column_headers,
    y_axis_column_header=y_axis_column_headers[0]
)
    else:
        trace_code = """
# Add the scatter traces to the figure
for column_header in {y_axis_column_headers}:
    fig.add_trace(go.Scatter(
        x={df_name}['{x_axis_column_header}'],
        y={df_name}[column_header],
        mode='markers',
        name=column_header
    ))
""".format(
    df_name=df_name, 
    y_axis_column_headers=y_axis_column_headers,
    x_axis_column_header=x_axis_column_headers[0]
)

    x_axis_title, y_axis_title = get_graph_labels(x_axis_column_headers, y_axis_column_headers)
    graph_title = get_graph_title(x_axis_column_headers, y_axis_column_headers, filtered, SCATTER)


    return """{CREATE_FIG_CODE}
{filtered_code}
{trace_code}
# Update the layout
# See Plotly documentation for cutomizations: https://plotly.com/python/reference/scatter/
fig.update_layout(
    xaxis_title='{x_axis_title}',
    yaxis_title='{y_axis_title}',
    title='{graph_title}',
)
{SHOW_FIG_CODE}""".format(
    CREATE_FIG_CODE=CREATE_FIG_CODE,
    trace_code=trace_code,
    filtered_code=filtered_code,
    x_axis_title=x_axis_title, 
    y_axis_title=y_axis_title,
    graph_title=graph_title,
    SHOW_FIG_CODE=SHOW_FIG_CODE
)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
The Mito package, which contains functions for creating a Mito sheet. 

To generate a new sheet, simply run:

import mitosheet
mitosheet.sheet()

If running mitosheet.sheet() just prints text that looks like `MitoWidget(...`, then you need to 
install the JupyterLab extension manager by running:

jupyter labextension install @jupyter-widgets/jupyterlab-manager@2;

Run this command in the terminal where you installed Mito. It should take 5-10 minutes to complete.

Then, restart your JupyterLab instance, and refresh your browser. Mito should now render.

NOTE: if you have any issues with installation, please email book a demo time at https://hubs.ly/H0FL1920
"""

from mitosheet.mito_analytics import log
import os
import pandas as pd
from mitosheet.user import initialize_user
from mitosheet.mito_widget import MitoWidget, sheet
from mitosheet.saved_analyses import register_analysis
from mitosheet.errors import CreationError, EditError
from mitosheet._version import __version__
from mitosheet.utils import make_valid_header

# Export all the sheet functions
from mitosheet.sheet_functions import *
# And the functions for changing types
from mitosheet.sheet_functions.types import *

# Export the graph filtering code, so that it can be used in the code that generates graphs
from mitosheet.api.graph.graph_utils import filter_df_to_safe_size

# Make sure the user is initalized
initialize_user()
log('imported_mitosheet', check_mitosheet_install=True)

# This function is only necessary for mitosheet3, as it is used
# in jlab3 to find the extension. It is not used in jlab2
def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": 'mitosheet3'
    }]




from typing import List
import pandas as pd
import plotly.graph_objects as go

from mitosheet.mito_analytics import log
from mitosheet.sheet_functions.types.utils import get_mito_type
from mitosheet.api.graph.histogram import get_histogram, get_histogram_code
from mitosheet.api.graph.graph_utils import BAR, BAR_CHART_MAX_NUMBER_OF_ROWS, CREATE_FIG_CODE, GRAPH_FILTER_LABELS, SHOW_FIG_CODE, X, Y, filter_df_to_safe_size, get_graph_labels, get_graph_title, get_starting_ranges



def get_bar_chart(df, x_axis_column_headers, y_axis_column_headers):
    """
    Returns a bar chart using the following heuristic:
    - Graphs the first BAR_CHART_MAX_NUMBER_OF_ROWS rows. 

    If only one axis has data, then we create a histogram instead. 

    If at leastat least the x_axis_column_headers or y_axis_column_headers 
    has at least one series, then it returns a bar chart. Otherwise, it returns 
    a blank graph.
    """
    x_axis = len(x_axis_column_headers) != 0
    y_axis = len(y_axis_column_headers) != 0
    
    # If no data is passed, return a blank graph
    if not x_axis and not y_axis:
        # This should never happen because this check already occurs in the get_graph
        # function. But we leave it here for robustness of the function
        return go.Figure(), False

    # If only one axis is supplied, then we create a histogram instead of a bar chart 
    # because it is more informative to show the counts of each value instead of graphing 
    # the values by index. 
    if x_axis and not y_axis:
        return get_histogram(X, df, x_axis_column_headers)
    elif y_axis and not x_axis:
        return get_histogram(Y, df, y_axis_column_headers)

    df, filtered = filter_df_to_safe_size(BAR, df, x_axis_column_headers, y_axis_column_headers)

    # Determine if the dataframe was sorted
    filter_label = GRAPH_FILTER_LABELS[BAR] if filtered else ''

    fig = go.Figure()

    # If the y_axis has __exactly__ one series and the x_axis has __at least__ one series (covered by above checks)
    if len(y_axis_column_headers) == 1:
        for column_header in x_axis_column_headers:
            fig.add_trace(
                go.Bar( 
                    x=df[column_header],
                    y=df[y_axis_column_headers[0]],
                    name=(' ').join([column_header, filter_label])
                )
            )
    # If the y_axis has > one series and the x_axis has __at least__ one series (covered by above checks)
    else: 
        for column_header in y_axis_column_headers:
            fig.add_trace(
                go.Bar( 
                    x=df[x_axis_column_headers[0]],
                    y=df[column_header],
                    name=(' ').join([column_header, filter_label])
                )
            )
        
    # Update the layout of the graph
    x_axis_title, y_axis_title = get_graph_labels(x_axis_column_headers, y_axis_column_headers)
    graph_title = get_graph_title(x_axis_column_headers, y_axis_column_headers, filtered, BAR)
    fig.update_layout(
        xaxis_title=x_axis_title,
        yaxis_title=y_axis_title,
        title=graph_title,
        barmode='group',
    )

    # Update the starting range of the graph
    ranges = get_starting_ranges(BAR, df, x_axis_column_headers, y_axis_column_headers)
    fig.update_layout(
        xaxis=dict(
            range=ranges[X]
        ), 
        yaxis=dict(
            range=ranges[Y]
        )
    )
    # Determine if the graph was zoomed in
    is_zoomed = ranges[X][0] is not None or ranges[X][1] or ranges[Y][0] is not None or ranges[Y][1] is not None

    log(f'generate_graph', {
        'params_graph_type': BAR,
        'params_x_axis_column_headers': x_axis_column_headers,
        'params_x_axis_column_types': [get_mito_type(df[column_header]) for column_header in x_axis_column_headers] if x_axis_column_headers is not None else [],
        'params_y_axis_column_headers': y_axis_column_headers,
        'params_y_axis_column_types': [get_mito_type(df[column_header]) for column_header in y_axis_column_headers] if y_axis_column_headers is not None else [],
        'params_filtered': filtered,
        'params_is_zoomed': is_zoomed
    })

    return fig, is_zoomed

def get_bar_code(
        df: pd.DataFrame, 
        x_axis_column_headers: List[str], 
        y_axis_column_headers: List[str],
        df_name: str
    ):
    """
    Generates the code for a bar graph, in the same way that the bar is
    created above (leaving out zooming).
    """

    x_axis = len(x_axis_column_headers) != 0
    y_axis = len(y_axis_column_headers) != 0

    if x_axis and not y_axis:
        return get_histogram_code(X, df, x_axis_column_headers, df_name)
    elif y_axis and not x_axis:
        return get_histogram_code(Y, df, y_axis_column_headers, df_name)

    _, filtered = filter_df_to_safe_size(BAR, df, x_axis_column_headers + y_axis_column_headers)
    filtered_code = ''
    if filtered:
        filtered_code = """
# Filter the dataframe so that it does not crash the browser
from mitosheet import filter_df_to_safe_size
{df_name}_filtered, _ = filter_df_to_safe_size('bar', {df_name}, {x_axis_column_headers}, {y_axis_column_headers})
""".format(
    df_name=df_name,
    x_axis_column_headers=x_axis_column_headers,
    y_axis_column_headers=y_axis_column_headers
)
        df_name = f'{df_name}_filtered'



    if len(y_axis_column_headers) == 1:
        trace_code = """
# Add the bar chart traces to the graph
for column_header in {x_axis_column_headers}:
    fig.add_trace(
        go.Bar( 
            x={df_name}[column_header],
            y={df_name}['{y_axis_column_header}'],
            name=column_header
        )
    )
""".format(
        x_axis_column_headers=x_axis_column_headers,
        y_axis_column_header=y_axis_column_headers[0],
        df_name=df_name
    )
        
    # If the y_axis has > one series and the x_axis has __at least__ one series (covered by above checks)
    else: 
        trace_code = """
# Add the bar chart traces to the graph
for column_header in {y_axis_column_headers}:
    fig.add_trace(
        go.Bar( 
            x={df_name}[column_header],
            y={df_name}['{x_axis_column_heade}'],
            name=column_header
        )
    )
""".format(
        y_axis_column_headers=y_axis_column_headers,
        x_axis_column_header=x_axis_column_headers[0],
        df_name=df_name
    )

    # Update the layout of the graph
    x_axis_title, y_axis_title = get_graph_labels(x_axis_column_headers, y_axis_column_headers)
    graph_title = get_graph_title(x_axis_column_headers, y_axis_column_headers, filtered, BAR)

    return """{CREATE_FIG_CODE}
{filtered_code}
{trace_code}
# Update the title and stacking mode of the graph
# See Plotly documentation for customizations: https://plotly.com/python/reference/bar/
fig.update_layout(
    xaxis_title='{x_axis_title}',
    yaxis_title='{y_axis_title}',
    title='{graph_title}',
    barmode='group',
)
{SHOW_FIG_CODE}""".format(
    CREATE_FIG_CODE=CREATE_FIG_CODE,
    filtered_code=filtered_code,
    trace_code=trace_code,
    x_axis_title=x_axis_title,
    y_axis_title=y_axis_title,
    graph_title=graph_title,
    SHOW_FIG_CODE=SHOW_FIG_CODE
)




def get_column_dtype(event, steps_manager):
    """
    Sends back the dtype of the column
    """
    sheet_index = event['sheet_index']
    column_header = event['column_header']

    series = steps_manager.dfs[sheet_index][column_header]

    return str(series.dtype)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import copy
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.state import DATAFRAME_SOURCE_DUPLICATED, State
from typing import Any, Dict, List, Optional


class DataframeDuplicateStepPerformer(StepPerformer):
    """
    This steps duplicates a dataframe of a given index. 
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'dataframe_duplicate'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Duplicated a Dataframe'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'dataframe_duplicate_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        **params
    ) -> State:
        post_state = copy(prev_state)

        # Execute the step
        df_copy = post_state.dfs[sheet_index].copy(deep=True)
        new_name = get_first_unused_name(post_state.df_names, post_state.df_names[sheet_index] + '_copy')
        post_state.add_df_to_state(df_copy, DATAFRAME_SOURCE_DUPLICATED, df_name=new_name)

        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int
    ) -> List[str]:
        old_df_name = post_state.df_names[sheet_index]
        new_df_name = post_state.df_names[len(post_state.dfs) - 1]

        return [f'{new_df_name} = {old_df_name}.copy(deep=True)']

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            old_df_name = df_names[sheet_index]
            new_df_name = df_names[len(df_names) - 1]
            return f'Duplicated {old_df_name} to {new_df_name}'
        return f'Duplicated a df'


def get_first_unused_name(df_names, dataframe_name):
    """
    Appends _1, _2, .. to df name until it finds an unused 
    dataframe name. If no append is necessary, will just return
    """
    if dataframe_name not in df_names:
        return dataframe_name

    for i in range(len(df_names) + 1):
        new_name = f'{dataframe_name}_{i + 1}'
        if new_name not in df_names:
            return new_name


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains handlers for the Mito API
"""
from mitosheet.api.get_pivot_params import get_pivot_params
from queue import Queue
from threading import Thread

from mitosheet.mito_analytics import log_event_processed
from mitosheet.api.graph import get_graph
from mitosheet.api.get_path_join import get_path_join
from mitosheet.api.get_datafiles import get_datafiles
from mitosheet.api.get_path_contents import get_path_contents
from mitosheet.api.get_column_dtype import get_column_dtype
from mitosheet.api.get_import_summary import get_import_summary
from mitosheet.api.get_column_describe import get_column_describe
from mitosheet.api.get_dateframe_as_csv import get_dataframe_as_csv
from mitosheet.api.get_saved_analysis_description import get_saved_analysis_description


# As the column summary statistics tab does three calls, we defaulted to this max
MAX_QUEUED_API_CALLS = 3

# NOTE: BE CAREFUL WITH THIS. When in development mode, you can set it
# so the API calls are handled in the main thread, to make printing easy
THREADED = True

class API():
    """
    The API provides a wrapper around a thread that responds to API calls.

    Some notes:
    -   We allow at most MAX_QUEUED_API_CALLS API calls to be in the queue, which practically
        Stops a backlog of calls from building up.
    -   All API calls should only be reads. This stops us from having to worry 
        about most concurrency issues
    -   Note that printing inside of a thread does not work properly! Use sys.stdout.flush() after the print statement.
        See here: https://stackoverflow.com/questions/18234469/python-multithreaded-print-statements-delayed-until-all-threads-complete-executi
    """
    def __init__(self, steps_manager, send):
        self.api_queue = Queue(MAX_QUEUED_API_CALLS)
        # Note that we make the thread a daemon thread, which practically means that when
        # The process that starts this thread terminate, our API will terminate as well.
        self.thread = Thread(target=handle_api_event_thread, args=(self.api_queue, steps_manager, send), daemon=True)
        self.thread.start()

        # Save some variables for ease
        self.steps_manager = steps_manager
        self.send = send

    def process_new_api_call(self, event):
        """
        We privilege new API calls over old calls, and evict the old ones 
        if the API queue is full.

        Because we are using a queue, only events that have not been started 
        being processed will get removed.
        """        
        if THREADED:
            if self.api_queue.full():
                self.api_queue.get()
            self.api_queue.put(event)
        else:
            handle_api_event(self.send, event, self.steps_manager)



def handle_api_event_thread(queue, steps_manager, send):
    """
    This is the worker thread function, that actually is 
    responsible for handling at the API call events.

    It lives forever, and just handles events as it 
    receives them from the queue
    """
    while True:
        # Note that this blocks when there is nothing in the queue,
        # and waits till there is something there - so no infinite 
        # loop as it is waiting!
        event = queue.get()
        # We place the API handling inside of a try catch, 
        # because otherwise if an error is thrown, then the entire thread crashes, 
        # and then the API never works again
        try:
            handle_api_event(send, event, steps_manager)
        except:
            # Log in error if it occurs
            log_event_processed(event, steps_manager, failed=True)

def handle_api_event(send, event, steps_manager):
    """
    Handler for all API calls. Note that any response to the
    API must return the same ID that the incoming message contains,
    so that the frontend knows how to match the responses.
    """    

    # And then handle it
    if event['type'] == 'datafiles':
        result = get_datafiles(event)
    elif event['type'] == 'get_path_contents':
        result = get_path_contents(event)
    elif event['type'] == 'get_path_join':
        result = get_path_join(event)
    elif event['type'] == 'import_summary':
        result = get_import_summary(event)
    elif event['type'] == 'get_dataframe_as_csv':
        result = get_dataframe_as_csv(event, steps_manager)
    elif event['type'] == 'get_graph':
        result = get_graph(event, steps_manager)
    elif event['type'] == 'get_column_describe':
        result = get_column_describe(event, steps_manager)
    elif event['type'] == 'get_column_dtype':
        result = get_column_dtype(event, steps_manager)
    elif event['type'] == 'get_saved_analysis_description':
        result = get_saved_analysis_description(event, steps_manager)
    elif event['type'] == 'get_pivot_params':
        result = get_pivot_params(event, steps_manager)
    else:
        raise Exception(f'Event: {event} is not a valid API call')
    
    send({
        'event': 'api_response',
        'id': event['id'],
        'data': result
    })



def get_dataframe_as_csv(event, steps_manager):
    """
    Sends a dataframe as a CSV string
    """
    sheet_index = event['sheet_index']
    df = steps_manager.dfs[sheet_index]

    return df.to_csv(index=False)


"""
Contains tests for the user utils, which determine the current
state of the user (e.g. should they upgrade).
"""
import pytest
from datetime import datetime, timedelta
import os

from mitosheet.user.schemas import USER_JSON_VERSION_1, USER_JSON_VERSION_2
from mitosheet.user.db import USER_JSON_PATH
from mitosheet.user import initialize_user, is_local_deployment, is_on_kuberentes_mito, is_running_test, should_upgrade_mitosheet
from mitosheet.tests.user.conftest import write_fake_user_json, today_str


def test_is_local():
    assert is_local_deployment()
    assert not is_on_kuberentes_mito()

def test_detects_tests():
    assert is_running_test()


def test_should_not_upgrade_on_first_creation():
    initialize_user()
    assert not should_upgrade_mitosheet()
    os.remove(USER_JSON_PATH)


def test_should_prompt_upgrade_after_21_days():
    write_fake_user_json(
        USER_JSON_VERSION_2,
        mitosheet_last_upgraded_date=(datetime.today() - timedelta(days=20)).strftime('%Y-%m-%d'),
    )

    initialize_user()
    assert not should_upgrade_mitosheet()

    write_fake_user_json(
        USER_JSON_VERSION_1,
        mitosheet_last_upgraded_date=(datetime.today() - timedelta(days=22)).strftime('%Y-%m-%d'),
    )

    initialize_user()
    assert should_upgrade_mitosheet()
    os.remove(USER_JSON_PATH)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
File that holds all the type conversion functions
"""

from mitosheet.sheet_functions.types.utils import (
    BOOLEAN_SERIES,
    DATETIME_SERIES,
    TIMEDELTA_SERIES,
    NUMBER_SERIES,
    STRING_SERIES,
)

from mitosheet.sheet_functions.types.to_boolean_series import to_boolean_series
from mitosheet.sheet_functions.types.to_datetime_series import to_datetime_series
from mitosheet.sheet_functions.types.to_number_series import to_number_series
from mitosheet.sheet_functions.types.to_string_series import to_string_series

SERIES_CONVERSION_FUNCTIONS = {
    BOOLEAN_SERIES: to_boolean_series,
    DATETIME_SERIES: to_datetime_series,
    NUMBER_SERIES: to_number_series,
    STRING_SERIES: to_string_series,
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Exports an error box that can be used to pass the errors we need. 

Further exports helper functions for creating errors
of different flavors. We wrap these functions in the same
EditError box to avoid creating too many new classes!

See more about why we use errors here: 
https://stackoverflow.com/questions/16138232/is-it-a-good-practice-to-use-try-except-else-in-python
"""
import traceback
from typing import Set, List

class CreationError(Exception):
    """
    An error that occurs during a mito.sheet call, on creation
    of the mito widget. 
    """
    def __init__(self, type_, header, to_fix):
        """
        Creates a creation error. 

        type_: the error id string.
        header: the header of the error message to be displayed.
        to_fix: instructions on how to handle or fix the error.
        """
        self.type_ = type_ # we have an _ to avoid overwriting the build in type
        self.header = header
        self.to_fix = to_fix

class EditError(Exception):
    """
    An error that occurs during the processing of an editing event.
    """
    def __init__(self, type_, header, to_fix):
        """
        Creates a creation error. 

        type_: a string that is the error type. 
        header: the header of the error message to be displayed.
        to_fix: instructions on how to handle or fix the error.
        """
        self.type_ = type_ # we have an _ to avoid overwriting the build in type
        self.header = header
        self.to_fix = to_fix

def make_no_sheet_error(sheet_indexes: Set[str]):
    """
    Helper function for creating a no_sheet_error.

    Occurs when a user edits a formula with a reference to a sheet that does not exist.
    """
    to_fix = f'There is an issue behind the scenes with that operation, we\'ll get it fixed as soon as possible!'

    return EditError(
        'no_sheet_error', 
        'No Sheet Exists',
        to_fix
    )

def make_no_column_error(column_headers: Set[str]):
    """
    Helper function for creating a no_column_error.

    Occurs when a user edits a formula with a reference to a column that does not exist.
    """
    if len(column_headers) == 1:
        to_fix = f'Sorry, there is no column with the name {next(iter(column_headers))}. Did you type it correctly?'
    else:
        to_fix = f'Sorry, there are no column with the names {", ".join(column_headers)}. Did you type them correctly?'

    return EditError(
        'no_column_error', 
        'No Column Exists',
        to_fix
    )

def make_column_exists_error(column_header: str):
    """
    Helper function for creating a column_exists_error.

    Occurs when:
    -  the user adds a column that already exists in the dataframe.
    """
    return EditError(
        'column_exists_error', 
        'Column Already Exists',
        f'Sorry, a column already exists with the name {column_header}. Try picking a different name!'
    )

def make_invalid_formula_error(formula: str, to_fix=None):
    """
    Helper function for creating a invalid_formula_error.

    Occurs when:
    -  the user edits a cell to an expression that cannot be parsed.
    """
    if to_fix is None:
        to_fix = f'Sorry, the formula \'{formula}\' is invalid. Did you type it correctly?'

    return EditError(
        'invalid_formula_error',
        'Invalid Formula',
        to_fix
    )

def make_circular_reference_error():
    """
    Helper function for creating a circular_reference_error.

    Occurs when:
    -  an edit introduces a circular reference.
    """
    return EditError(
        'circular_reference_error',
        'Circular Reference',
        f'Sorry, circular references are not supported currently.'
    )

def make_wrong_column_metatype_error(column_header: str):
    """
    Helper function for creating a wrong_column_metatype_error.

    Occurs when:
    -  a user edits a column to a formula when it is not a formula type column.
    """
    return EditError(
        'wrong_column_metatype_error',
        'Wrong Column Type',
        f'Sorry, the column {column_header} is a data column. We don\'t currently support changing it to a formula.'
    )

def make_invalid_column_headers_error(column_headers: List[str]):
    """
    Helper function for creating a invalid_column_headers_error.

    Occurs when:
    -  a user creates (or renames) a column(s) that has an invalid name.
    """
    to_fix = f'All headers in the dataframe must contain at least one letter and no symbols other than numbers and "_". Invalid headers: {", ".join(column_headers)}'

    return EditError(
        'invalid_column_header_error',
        'Invalid Column Header',
        to_fix
    )

def make_execution_error():
    """
    Helper function for creating a execution_error.

    Occurs when:
    -  running an existing analysis on the parameters passed to the mito sheet throw an error.
    """
    # TODO: this to_fix message is _just terrible_
    return EditError(
        'execution_error',
        'Execution Error',
        f'Sorry, there was an error during executing this code.'
    )

def make_function_execution_error(function):
    """
    Helper function for creating a function_execution_error.

    Occurs when:
    -  An error occurs inside of a mito sheet function
    """
    # TODO: this to_fix message is _just terrible_
    return EditError(
        'execution_error',
        f'Error Executing {function}',
        f'Sorry, there was an error in the {function}. Please check the documentation to make sure you called it correctly.'
    )

def make_unsupported_function_error(functions: Set[str]):
    """
    Helper function for creating a unsupported_function_error.

    Occurs when:
    -  the user's new formula containing a function that is not supported by mito.
    """
    # TODO: we could fix up this to_fix, lol.
    if len(functions) == 1:
        to_fix = f'Sorry, mito does not currently support the function {next(iter(functions))}. To request it, shoot us an email!'
    else:
        to_fix = f'Sorry, mito does not currently support the functions {", ".join(functions)}. To request it, shoot us an email!'

    return EditError(
        'unsupported_function_error',
        'Unsupported Function',
        to_fix
    )

def make_invalid_column_delete_error(column_header, dependents):
    """
    Helper function for creating a invalid_column_delete_error.

    Occurs when:
    -  the user deletes a column that is referenced by other columns
    """
    # We make sure it's a list, for easy accessing!
    dependents = list(dependents)

    # We make the error message read nicely!
    if len(dependents) == 1:
        dependent_string = f'column {dependents[0]}'
    elif len(dependents) == 2:
        dependent_string = f'columns {dependents[0]} and {dependents[1]}'
    else:
        dependent_string = f'columns {", ".join(dependents[: -1])}, and {dependents[-1]}'

    return EditError(
        'invalid_column_delete_error',
        'Column Has Dependents',
        f'{column_header} cannot be deleted, as it is referenced in {dependent_string}. Please remove these references before deleting.'
    )

def make_invalid_arguments_error(function):
    """
    Helper function for creating a invalid_arguments_error.

    Occurs when:
    -  the user calls a sheet function with incorrect arguments
    """
    return EditError(
        'invalid_arguments_error',
        f'Invalid Arguments to {function}',
        f'It looks like you passed the wrong arguments to {function}. Try checking out the documentation to see correct usage!'
    )

def make_invalid_aggregation_error():
    """
    Helper function for creating a invalid_aggregation_error.

    Occurs when:
    -  the user tries to .agg with a function on a column that has the wrong type (e.g. 'mean' on a
        column of strings).
    """
    return EditError(
        'invalid_aggregation_error',
        f'Invalid Aggregation Function',
        f'Sorry, you tried to aggregate with a function that could not be applied to that type of column. Please try again!'
    )

def make_invalid_filter_error(filter_value, correct_type):
    """
    Helper function for creating a invalid_filter_error.

    Occurs when:
    - A user tries to create a filter with an invalid value (e.g. a string, when filtering numbers
      or a non-date when filtering dates).
    """

    if correct_type == 'datetime':
        correct_format = 'as YYYY-MM-DD.'
    elif correct_type == 'number':
        correct_format = 'that is a number.'
    else:
        correct_format = 'that is valid.'

    return EditError(
        'invalid_filter_error',
        f'Invalid Filter',
        f'Sorry, the value {filter_value} is not a valid value for that {correct_type} filter. Please enter a value {correct_format}!'
    )

def make_invalid_sort_error(column_header):
    """
    Helper function for creating a invalid_sort_error.

    Occurs when:
    - A user tries to sort a dataframe on a column of mixed types 
    (e.g. a column with strings and floats).
    """

    return EditError(
        'invalid_sort_error',
        f'Invalid Sort',
        f'Sorry, the column {column_header} has mixed data types. Please make sure the column has one datatype before trying to sort.'
    )

def make_df_exists_error(df_name):
    """
    Helper function for creating a df_exists_error.

    Occurs when:
    - A user tries to create a dataframe with the name of an already existing
    dataframe in the sheet.
    """

    return EditError(
        'df_exists_error',
        f'Dataframe Exists',
        f'Sorry, the dataframe {df_name} already exists. Please pick a different name.'
    )

def make_invalid_column_type_change_error(column_header, old_dtype, new_dtype):
    """
    Helper function for creating a invalid_column_type_change_error.

    Occurs when:
    - A user tries to change the type of one column to an incompatible type (e.g. from
      a number to a datetime, which is impossible).
    """

    return EditError(
        'invalid_column_type_change_error',
        f'Invalid Type Change',
        f'Sorry, the column {column_header} has a type {old_dtype}, which cannot be changed to the type {new_dtype}.'
    )

def make_invalid_pivot_error():
    """
    Helper function for creating a invalid_pivot_error.

    Occurs when:
    -  the user runs a pivot that is invalid in some way.
    """
    # TODO: this to_fix message is _just terrible_
    return EditError(
        'invalid_pivot_error',
        'Pivot Error',
        f'Sorry, there was an error computing your pivot. Please try a different pivot!'
    )


def make_is_directory_error(file_name):
    """
    Helper function for creating directory_error

    Occurs when:
    - the user tries to import a directory in a simple import
    """
    return EditError(
        'directory_error',
        f'Change Path to a File',
        f'Sorry, the path {file_name} is a directory. Please pass in a path to a file to import it into Mito.'
    )


ARG_FULL_NAME = {
    'int': 'number',
    'float': 'number',
    'str': 'string',
    'bool': 'boolean'
}

def make_operator_type_error(operator, arg_one_type, arg_two_type):
    """
    Helper function for creating a operator_type_error.

    Occurs when:
    - user uses an operator to add/mul/div (etc) invalid types. E.g. they do: 1 + "hi",
      or pd.Series(['hi']) / 10.
    """

    # Provide a helpful error message, if possible, for the case of a number and a string
    # (which is the most common of these errors).
    if (arg_one_type == 'str' and (arg_two_type == 'int' or arg_two_type == 'float')) or \
        (arg_two_type == 'str' and (arg_one_type == 'int' or arg_one_type == 'float')):
        to_fix = f'Sorry, you tried to {operator} on a string and a number. Please wrap the number argument in a call to the VALUE function.'
    else:
        # Otherwise, handle errors in a non-specific way - this is a catch all for the rest of operator type errors
        type_one_full_name = ARG_FULL_NAME[arg_one_type] if arg_one_type in ARG_FULL_NAME else arg_one_type
        type_two_full_name = ARG_FULL_NAME[arg_two_type] if arg_two_type in ARG_FULL_NAME else arg_two_type
        # TODO: once we have a step that changes the column type, we should update this error to be more helpful
        to_fix = f'Sorry, you tried to use {operator} on a {type_one_full_name} and a {type_two_full_name}. Please change the type of one of these columns to be compatible.'

    return EditError(
        'operator_type_error',
        f'Error with {operator}',
        to_fix
    )

def get_recent_traceback():
    return traceback.format_exc()

def get_recent_traceback_as_list():
    """
    Helper function for getting the traceback of the most recent error.
    
    This, in turns, gives us a stack trace for any error that we want
    to log with detail, in a format that is sendable to mixpanel.

    We get this as a list of strings, as mixpanel truncates strings at 255 chars
    and thus we avoid things getting chopped this way.
    """
    # NOTE: We ignore empty lines, as they add no information
    return [line for line in traceback.format_exc().split('\n') if line != '']



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains all functions that can be used in a sheet that operate on
numbers.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
import functools
import pandas as pd
import numpy as np

from mitosheet.sheet_functions.types import NUMBER_SERIES
from mitosheet.sheet_functions.types.decorators import filter_nans, convert_args_to_series_type, convert_arg_to_series_type, handle_sheet_function_errors
from mitosheet.sheet_functions.sheet_function_utils import fill_series_list_to_max, fill_series_to_length

@handle_sheet_function_errors
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
def ABS(series: pd.Series):
    """
    {
        "function": "ABS",
        "description": "Returns the absolute value of the passed number or series.",
        "examples": [
            "ABS(-1.3)",
            "ABS(A)"
        ],
        "syntax": "ABS(value)",
        "syntax_elements": [{
                "element": "value",
                "description": "The value or series to take the absolute value of."
            }
        ]
    }
    """
    return series.abs()


@handle_sheet_function_errors
@filter_nans
@convert_args_to_series_type(NUMBER_SERIES, 'skip', ('default', np.NaN))
def AVG(*argv) -> pd.Series:
    """
    {
        "function": "AVG",
        "description": "Returns the numerical mean value of the passed numbers and series.",
        "examples": [
            "AVG(1, 2)",
            "AVG(A, B)",
            "AVG(A, 2)"
        ],
        "syntax": "AVG(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or series to consider when calculating the average."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or series to consider when calculating the average."
            }
        ]
    }
    """
    # Fill sum to the max length of any series
    argv = fill_series_list_to_max(argv)

    arg_sum = functools.reduce((lambda x, y: x + y), argv) 
    return arg_sum / len(argv)


@handle_sheet_function_errors
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
@convert_arg_to_series_type(
    1,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
def CORR(s1, s2) -> pd.Series:
    """
    {
        "function": "CORR",
        "description": "Computes the correlation between two series, excluding missing values.",
        "examples": [
            "=CORR(A, B)",
            "=CORR(B, A)"
        ],
        "syntax": "CORR(series_one, series_two)",
        "syntax_elements": [{
                "element": "series_one",
                "description": "The number series to convert to calculate the correlation."
            }, {
                "element": "series_two",
                "description": "The number series to convert to calculate the correlation."
            }
        ]
    }
    """
    corr = s1.corr(s2, method='pearson')
    return pd.Series(
        [corr] * len(s1)
    )


@handle_sheet_function_errors
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
# We don't use the @filter_nans decorator because kurtosis ignores NaN values
def KURT(series) -> pd.Series:
    """
    {
        "function": "KURT",
        "description": "Computes the unbiased kurtosis, a measure of tailedness, of a series, excluding missing values.",
        "examples": [
            "=KURT(A)",
            "=KURT(A * B)"
        ],
        "syntax": "KURT(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the unbiased kurtosis of."
            }
        ]
    }
    """
    kurtosis = series.kurtosis()
    return pd.Series(
        [kurtosis] * len(series)
    )


@handle_sheet_function_errors
@filter_nans
# NOTE: we set the default to -inf, so that we ignore values that fail to convert
@convert_args_to_series_type(NUMBER_SERIES, 'skip', ('default', float("-inf")))
def MAX(*argv) -> pd.Series:
    """
    {
        "function": "MAX",
        "description": "Returns the maximum value among the passed arguments.",
        "examples": [
            "MAX(10, 11)",
            "MAX(Old_Data, New_Data)"
        ],
        "syntax": "MAX(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to consider for the maximum value."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to compute the maximum value from."
            }
        ]
    }
    """
    # We make a dataframe out of the series, and take the max between all the columns
    return pd.DataFrame(argv).max()


@handle_sheet_function_errors
@filter_nans
# NOTE: we set the default to inf, so that we ignore values that fail to convert
@convert_args_to_series_type(NUMBER_SERIES, 'skip', ('default', float("inf")))
def MIN(*argv) -> pd.Series:
    """
    {
        "function": "MIN",
        "description": "Returns the minimum value among the passed arguments.",
        "examples": [
            "MIN(10, 11)",
            "MIN(Old_Data, New_Data)"
        ],
        "syntax": "MIN(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to consider for the minumum value."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to compute the minumum value from."
            }
        ]
    }
    """
    # We make a dataframe out of the series, and take the min between all the columns
    return pd.DataFrame(argv).min()


@handle_sheet_function_errors
@filter_nans
@convert_args_to_series_type(NUMBER_SERIES, 'skip', ('default', 1))
def MULTIPLY(*argv) -> pd.Series:
    """
    {
        "function": "MULTIPLY",
        "description": "Returns the product of two numbers.",
        "examples": [
            "MULTIPLY(2,3)",
            "MULTIPLY(A,3)"
        ],
        "syntax": "MULTIPLY(factor1, [factor2, ...])",
        "syntax_elements": [{
                "element": "factor1",
                "description": "The first number to multiply."
            },
            {
                "element": "factor2, ... [OPTIONAL]",
                "description": "Additional numbers or series to multiply."
            }
        ]
    }
    """
    # We make sure all the series are the max length, so the * has something at every index
    argv = fill_series_list_to_max(argv)

    return functools.reduce((lambda x, y: x * y), argv) 


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
@convert_arg_to_series_type(
    1,
    NUMBER_SERIES,
    'skip',
    ('default', 1)
)
def POWER(series, power):
    """
    {
        "function": "POWER",
        "description": "The POWER function can be used to raise a number to a given power.",
        "examples": [
            "POWER(4, 1/2)",
            "POWER(Dose, 2)"
        ],
        "syntax": "POWER(value, exponent)",
        "syntax_elements": [{
                "element": "value",
                "description": "Number to raise to a power."
            },
            {
                "element": "exponent",
                "description": "The number to raise value to."
            }
        ]
    }
    """
    power = fill_series_to_length(power, len(series))
    return series.pow(power)


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
@convert_arg_to_series_type(
    1,
    NUMBER_SERIES,
    'error',
    ('default', 2),
    optional=True
)
def ROUND(series, decimals=None):
    """
    {
        "function": "ROUND",
        "description": "Rounds a number to a given number of decimals.",
        "examples": [
            "ROUND(1.3)",
            "ROUND(A, 2)"
        ],
        "syntax": "ROUND(value, [decimals])",
        "syntax_elements": [{
                "element": "value",
                "description": "The value or series to round."
            },
            {
                "element": "decimals",
                "description": " The number of decimals to round to. Default is 0."
            }
        ]
    }
    """

    # If no decimals option is passed, round to no decimals
    if decimals is None:
        return series.round()
    
    # Otherwise, fill the decimals to length
    decimals = fill_series_to_length(decimals, series.size)

    return pd.Series(
        [round(num, dec) for num, dec in zip(series, decimals)]
    )

@handle_sheet_function_errors
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
# We don't use the @filter_nans decorator because skew ignores NaN values
def SKEW(series) -> pd.Series:
    """
    {
        "function": "SKEW",
        "description": "Computes the skew of a series, excluding missing values.",
        "examples": [
            "=SKEW(A)",
            "=SKEW(A * B)"
        ],
        "syntax": "SKEW(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the skew of."
            }
        ]
    }
    """
    skew = series.skew()
    return pd.Series(
        [skew] * len(series)
    )


@handle_sheet_function_errors
@filter_nans
@convert_args_to_series_type(NUMBER_SERIES, 'skip', ('default', 0))
def SUM(*argv) -> pd.Series:
    """
    {
        "function": "SUM",
        "description": "Returns the sum of the given numbers and series.",
        "examples": [
            "SUM(10, 11)",
            "SUM(A, B, D, F)",
            "SUM(A, B, D, F)"
        ],
        "syntax": "SUM(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to add together."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to sum."
            }
        ]
    }
    """
    # Fill sum to the max length of any series
    argv = fill_series_list_to_max(argv)

    return functools.reduce((lambda x, y: x + y), argv) 


@handle_sheet_function_errors
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
# We don't use the @filter_nans decorator because std ignores NaN values
def STDEV(series) -> pd.Series:
    """
    {
        "function": "STDEV",
        "description": "Computes the standard deviation of a series, excluding missing values.",
        "examples": [
            "=STDEV(A)",
            "=STDEV(A * B)"
        ],
        "syntax": "STDEV(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the standard deviation of."
            }
        ]
    }
    """
    std = series.std()
    return pd.Series(
        [std] * len(series)
    )



@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
def VALUE(series) -> pd.Series:
    """
    {
        "function": "VALUE",
        "description": "Converts a string series to a number series. Any values that fail to convert will return an NaN.",
        "examples": [
            "=VALUE(A)",
            "=VALUE('123')"
        ],
        "syntax": "VALUE(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series to convert to a number."
            }
        ]
    }
    """
    return series


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
def EXP(series) -> pd.Series:
    """
    {
        "function": "EXP",
        "description": "Returns e, the base of the natural logarithm, raised to the power of passed series.",
        "examples": [
            "=EXP(data)",
            "=EXP(A)"
        ],
        "syntax": "EXP(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to raise e to."
            }
        ]
    }
    """
    return np.exp(series)


@handle_sheet_function_errors
@convert_arg_to_series_type(
    0,
    NUMBER_SERIES,
    'error',
    ('default', np.NaN)
)
# We don't use the @filter_nans decorator because var ignores NaN values
def VAR(series) -> pd.Series:
    """
    {
        "function": "VAR",
        "description": "Computes the variance of a series, excluding missing values.",
        "examples": [
            "=VAR(A)",
            "=VAR(A - B)"
        ],
        "syntax": "VAR(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the variance of."
            }
        ]
    }
    """
    var = series.var()
    return pd.Series(
        [var] * len(series)
    )

# TODO: we should see if we can list these automatically!
NUMBER_FUNCTIONS = {
    'ABS': ABS,
    'AVG': AVG,
    'CORR': CORR,
    'EXP': EXP,
    'KURT': KURT,
    'MAX': MAX,
    'MIN': MIN,
    'MULTIPLY': MULTIPLY,
    'POWER': POWER,
    'ROUND': ROUND,
    'SKEW': SKEW,
    'SUM': SUM,
    'STDEV': STDEV,
    'VALUE': VALUE,
    'VAR': VAR
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.
from copy import deepcopy
from mitosheet.state import State
from typing import Any, Dict, List, Optional
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_sort_error
)

# CONSTANTS USED IN THE SORT STEP ITSELF
ASCENDING = 'ascending'
DESCENDING = 'descending'

class SortStepPerformer(StepPerformer):
    """
    Allows you to sort a df based on key column, in either
    ascending or descending order.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'sort'
    
    @classmethod
    def step_display_name(cls) -> str:
        return 'Sorted a Column'

    @classmethod
    def step_event_type(cls) -> str:
        return 'sort_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        column_header: str,
        sort_direction: str,
        **params
    ):
        """
        Returns the new new post state after sorting the sheet
        at `sheet_index` by the passed `column_header` in the given
        `sort_direction`
        """

        # if the sheet doesn't exist, throw an error
        if not prev_state.does_sheet_index_exist_within_state(sheet_index):
            raise make_no_sheet_error(sheet_index)

        # We check that the sorted column exists 
        missing_column = set([column_header]).difference(prev_state.column_metatype[sheet_index].keys())
        if len(missing_column) > 0: 
            raise make_no_column_error(missing_column)

        # We make a new state to modify it
        post_state = deepcopy(prev_state)

        try: 
            new_df = prev_state.dfs[sheet_index].sort_values(by=column_header, ascending=(sort_direction == ASCENDING), na_position='first')
            new_df = new_df.reset_index(drop=True)
            post_state.dfs[sheet_index] = new_df
        except TypeError as e:
            # A NameError occurs when you try to sort a column with incomparable 
            # dtypes (ie: a column with strings and floats)
            print(e)
            # Generate an error informing the user
            raise make_invalid_sort_error(column_header)

        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        column_header: str,
        sort_direction: str
    ) -> List[str]:
        df_name = post_state.df_names[sheet_index]
        return [
            f'{df_name} = {df_name}.sort_values(by=\'{column_header}\', ascending={sort_direction == ASCENDING}, na_position=\'first\')', 
            f'{df_name} = {df_name}.reset_index(drop=True)'
        ]

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        column_header: str,
        sort_direction: str,
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            df_name = df_names[sheet_index]
            return f'Sorted {column_header} in {df_name} in {sort_direction} order'
        return f'Sorted {column_header} in {sort_direction} order'


import os
from pathlib import Path


def get_filenames_with_suffix(*suffixes):
    """
    Returns all the file names in the current folder that end with the given
    suffix, sorted from most-recently created to the oldest.
    """
    # We sort them by creation time, to get the most recent files, as the user
    # is more likely to want these
    filenames = sorted(Path('.').iterdir(), key=os.path.getmtime)
    filenames.reverse()
    return [str(filename) for filename in filenames if filename.suffix in suffixes]


def get_datafiles(event):
    """
    Handles a `datafiles` api call, and returns all the csv files
    in the current folder.
    """
    csv_files = get_filenames_with_suffix(
        '.csv', 
        '.tsv',
        '.tab',
    )
    # TODO: also get the XLSX files, when we can import them
    return csv_files


# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

# TODO: we have to update the bump version script
import json
from pathlib import Path
import os
import tarfile
from typing import Union

def get_package_json():
    """
    This helper function is responsible for getting the package.json that is bundled
    with the javascript code bundle inside this Python package. 

    When the package is mitosheet (aka, jlab 2), then the mitosheet/labextension folder
    contains a .tgz zip file that contains the package.json. 

    When the package is mitosheet3 (aka, jlab3), then the mitosheet/labextension folder
    contains the package.json directly (as this is the new build process!).

    Thus, this function handles each case, to make sure we can get the current version 
    of the mitosheet package correctly.
    """
    lab_extension_folder = os.path.join(Path(__file__).parent, 'labextension')

    try:
        # First, try and get the jlab 3 case
        package_json_path = os.path.join(lab_extension_folder, 'package.json')
        return json.loads(open(package_json_path).read())
    except:
        # If we cannot get the package json directly, then it must be zipped (as it is
        # in jlab 2), and so we pull out the package json
        tar_file_name = list(filter(lambda x: x.endswith('tgz'), os.listdir(lab_extension_folder)))[0]
        tar_file_path = os.path.join(lab_extension_folder, tar_file_name)
        with tarfile.open(tar_file_path, 'r:gz') as t:
            package_json_file = t.extractfile('package/package.json')
            return json.loads(package_json_file.read())

__version__ = get_package_json()['version']